# Copyright (c) 2020 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import geopandas as gpd
import json
import pandas as pd
import numpy as np

from pyincore import Dataset, DataService, IncoreClient
from functools import reduce


class DataProcessUtil:

    @staticmethod
    def get_mapped_result_from_analysis(client, inventory_id: str, dmg_result_dataset,
                                        bldg_func_dataset, archetype_mapping_id: str,
                                        groupby_col_name: str = "max_state", arch_col="archetype"
                                        ):
        """Use this if you want to load results directly from the output files of the analysis, than storing the results
        to data service and loading from there using ids.
        It takes the static inputs: inventory & archetypes as dataset ids. The result inputs are taken as
        Dataset class objects, which are created by reading the output result files.

        Args:
            client: Service client with authentication info
            inventory_id: Inventory dataset id
            dmg_result_dataset: Incore dataset for damage result
            bldg_func_dataset: Incore dataset for building func dataset
            archetype_mapping_id: Mapping id dataset for archetype

        Returns:
            dmg_ret_json: JSON of the damage state results ordered by cluster and category.
            func_ret_json: JSON of the building functionality results ordered by cluster and category.
            mapped_df: Dataframe of max damage state
            arch_col: column name for the archetype to perform the merge

        """
        bldg_inv = Dataset.from_data_service(inventory_id, DataService(client))
        inventory = pd.DataFrame(gpd.read_file(bldg_inv.local_file_path))

        dmg_result = dmg_result_dataset.get_dataframe_from_csv()

        bldg_func_df = bldg_func_dataset.get_dataframe_from_csv()
        bldg_func_df.rename(columns={"building_guid": "guid", "samples": "failure"}, inplace=True)

        arch_mapping = Dataset.from_data_service(archetype_mapping_id, DataService(client)).get_dataframe_from_csv()

        max_state_df = DataProcessUtil.get_max_damage_state(dmg_result)
        dmg_ret_json = DataProcessUtil.create_mapped_dmg_result(inventory, max_state_df, arch_mapping,
                                                                groupby_col_name, arch_col)
        func_ret_json = DataProcessUtil.create_mapped_func_result(inventory, bldg_func_df, arch_mapping, arch_col)

        return dmg_ret_json, func_ret_json, max_state_df

    @staticmethod
    def get_mapped_result_from_dataset_id(client, inventory_id: str, dmg_result_id: str, bldg_func_id,
                                          archetype_mapping_id: str,
                                          groupby_col_name: str = "max_state", arch_col="archetype"):
        """Use this if your damage results are already stored in the data service and you have their dataset ids.
        All the inputs (except groupby_col_name) are dataset ids.

        Args:
            client: Service client with authentication info
            inventory_id: Inventory dataset id
            dmg_result_id: Damage result dataset id
            bldg_func_id: Incore dataset for building func id
            archetype_mapping_id: Mapping id dataset for archetype
            groupby_col_name: column name to group by, default to max_state
            arch_col: column name for the archetype to perform the merge

        Returns:
            dmg_ret_json: JSON of the damage state results ordered by cluster and category.
            func_ret_json: JSON of the building functionality results ordered by cluster and category.
            max_state_df: Dataframe of max damage state

        """
        bldg_inv = Dataset.from_data_service(inventory_id, DataService(client))
        inventory = pd.DataFrame(gpd.read_file(bldg_inv.local_file_path))

        dmg_result_dataset = Dataset.from_data_service(dmg_result_id, DataService(client))
        dmg_result = dmg_result_dataset.get_dataframe_from_csv()

        bldg_func_dataset = Dataset.from_data_service(bldg_func_id, DataService(client))
        bldg_func_df = bldg_func_dataset.get_dataframe_from_csv()
        bldg_func_df.rename(columns={"building_guid": "guid", "samples": "failure"}, inplace=True)

        archtype_mapping_dataset = Dataset.from_data_service(archetype_mapping_id, DataService(client))
        arch_mapping = archtype_mapping_dataset.get_dataframe_from_csv()

        max_state_df = DataProcessUtil.get_max_damage_state(dmg_result)
        dmg_ret_json = DataProcessUtil.create_mapped_dmg_result(inventory, max_state_df, arch_mapping,
                                                                groupby_col_name, arch_col)
        func_ret_json = DataProcessUtil.create_mapped_func_result(inventory, bldg_func_df, arch_mapping, arch_col)

        return dmg_ret_json, func_ret_json, max_state_df

    @staticmethod
    def get_mapped_result_from_path(inventory_path: str, dmg_result_path: str,
                                    func_result_path: str,
                                    archetype_mapping_path: str,
                                    groupby_col_name: str, arch_col="archetype"):
        """

        Args:
            inventory_path: Path to the zip file containing the inventory
                    example: /Users/myuser/5f9091df3e86721ed82f701d.zip
            dmg_result_path: Path to the damage result output file
            func_result_path: Path to the bldg functionality result output file
            archetype_mapping_path: Path to the arechetype mappings
            groupby_col_name: column name to group by, default to max_state
            arch_col: column name for the archetype to perform the merge

        Returns:
            dmg_ret_json: JSON of the damage state results ordered by cluster and category.
            func_ret_json: JSON of the building functionality results ordered by cluster and category.
            mapped_df: Dataframe of max damage state

        """
        inventory = pd.DataFrame(gpd.read_file("zip://" + inventory_path))
        dmg_result = pd.read_csv(dmg_result_path)
        bldg_func_df = pd.read_csv(func_result_path)
        bldg_func_df.rename(columns={"building_guid": "guid", "samples": "failure"}, inplace=True)
        arch_mapping = pd.read_csv(archetype_mapping_path)

        max_state_df = DataProcessUtil.get_max_damage_state(dmg_result)
        dmg_ret_json = DataProcessUtil.create_mapped_dmg_result(inventory, max_state_df, arch_mapping,
                                                                groupby_col_name, arch_col)

        func_ret_json = DataProcessUtil.create_mapped_func_result(inventory, bldg_func_df, arch_mapping, arch_col)

        return dmg_ret_json, func_ret_json, max_state_df

    @staticmethod
    def create_mapped_dmg_result(inventory, dmg_result, arch_mapping, groupby_col_name="max_state",
                                 arch_col="archetype"):
        """

        Args:
            inventory: dataframe represent inventory
            dmg_result: dmg_result that need to merge with inventory and be grouped
            arch_mapping: Path to the archetype mappings

        Returns:
            ret_json: JSON of the results ordered by cluster and category.

        """
        dmg_states = dmg_result[groupby_col_name].unique().tolist()  # get unique damage states
        dmg_merged = pd.merge(inventory, dmg_result, on="guid")
        mapped_df = pd.merge(dmg_merged, arch_mapping, on=arch_col)
        unique_categories = arch_mapping.groupby(by=["cluster", "category"], sort=False).count().reset_index()

        group_by = mapped_df.groupby(by=[groupby_col_name, "cluster", "category"]).count().reset_index()
        group_by = group_by.loc[:, ["guid", groupby_col_name, "cluster", "category"]]
        group_by.rename(columns={"guid": "count"}, inplace=True)

        pivot = group_by.pivot_table(values="count", index=["cluster", "category"], columns=groupby_col_name,
                                     fill_value=0)

        table = pd.DataFrame()
        table[["category", "cluster"]] = unique_categories[["category", "cluster"]]
        result_by_cluster = pd.merge(table, pivot, how="left", on=["cluster", "category"])

        # Add missing max damage states. Handles case when no inventory fall under some damage states.
        result_by_cluster = result_by_cluster.reindex(result_by_cluster.columns.union(
            dmg_states, sort=False), axis=1, fill_value=0)

        result_by_category = result_by_cluster.groupby(by=["category"], sort=False).sum(min_count=1).reset_index()

        result_by_cluster[dmg_states] = result_by_cluster[dmg_states].fillna(-1).astype(int)
        result_by_category[dmg_states] = result_by_category[dmg_states].fillna(-1).astype(int)

        cluster_records = result_by_cluster.to_json(orient="records")
        category_records = result_by_category.to_json(orient="records")
        json_by_cluster = json.loads(cluster_records)
        json_by_category = json.loads(category_records)

        return {"by_cluster": json_by_cluster, "by_category": json_by_category}

    @staticmethod
    def create_mapped_func_result(inventory, bldg_func, arch_mapping, arch_col="archetype"):
        """

        Args:
            inventory: dataframe represent inventory
            bldg_func: building func state dataset
            arch_mapping: Path to the archetype mappings
            arch_col: archetype column to use for the clustering

        Returns:
            ret_json: JSON of the results ordered by cluster and category.

        """
        def _sum_average(series):
            return reduce(lambda x, y: np.mean(x + y).round(0), series)

        func_state = ["percent_functional", "percent_non_functional", "num_functional", "num_non_functional"]

        # unify mcs and bldg func naming
        bldg_func.rename(columns={"building_guid": "guid", "samples": "failure"}, inplace=True)

        # drop nan but count their numbers
        count_null = (bldg_func["failure"] == "").sum()
        bldg_func = bldg_func[bldg_func['failure'] != ""]

        func_merged = pd.merge(inventory, bldg_func, on="guid")
        mapped_df = pd.merge(func_merged, arch_mapping, on=arch_col)
        unique_categories = arch_mapping.groupby(by=["category"], sort=False, as_index=False).count()["category"]
        unique_cluster = arch_mapping.groupby(by=["cluster", "category"], sort=False, as_index=False).count()[[
            "cluster", "category"]]

        mapped_df = mapped_df[["guid", "failure", "category", "cluster"]]
        mapped_df["failure_array"] = mapped_df["failure"].apply(lambda x: np.array([int(x) for x in x.split(",")]))

        def _group_by(by_column, unique):
            # group by cluster
            result = mapped_df.groupby(by=by_column, sort=False, as_index=False).agg(
                {"guid": "count", "failure_array": [_sum_average]})

            # clean up
            result.rename(columns={"guid": "tot_count", "failure_array": "num_functional"}, inplace=True)

            # 0 (failed), 1 (not failed). MCS
            # 0 otherwise (not functional), 1 (functional),  Functionality
            result["num_non_functional"] = result["tot_count"].squeeze() - result["num_functional"].squeeze()
            result["percent_functional"] = result["num_functional"].squeeze() / result["tot_count"].squeeze()
            result["percent_non_functional"] = 1 - result["percent_functional"]

            # remove the tuples in column
            result.columns = [x[0] if len(x) > 1 else x for x in result.columns]

            # more clean up
            result = pd.merge(unique, result, how="left", on=by_column)

            # Add missing max damage states. Handles case when no inventory fall under some damage states.
            result = result.reindex(result.columns.union(func_state, sort=False), axis=1, fill_value=0)

            # replace NaN
            result[func_state] = result[func_state].fillna(-1)
            result["tot_count"] = result["tot_count"].fillna(-1)
            result[["num_functional", "num_non_functional"]] = result[["num_functional", "num_non_functional"]].astype(int)

            return result

        result_by_cluster = _group_by(by_column=["cluster", "category"], unique=unique_cluster)
        result_by_category = _group_by(by_column=["category"], unique=unique_categories)

        cluster_records = result_by_cluster.to_json(orient="records")
        category_records = result_by_category.to_json(orient="records")
        json_by_cluster = json.loads(cluster_records)
        json_by_category = json.loads(category_records)

        return {"by_cluster": json_by_cluster, "by_category": json_by_category, "NA": int(count_null)}

    @staticmethod
    def get_max_damage_state(dmg_result):
        """Given damage result output decide the maximum damage state for each guid.

        Args:
            dmg_result (pd.DataFrame): damage result output, such as building damage, EPF damage and etc.

        Returns:
            pd.DataFrame: Pandas dataframe that has column GUID and column max_state.

        """
        if all(column in dmg_result.columns for column in ["DS_0", "DS_1", "DS_2", "DS_3"]):
            dmg_states = ["DS_0", "DS_1", "DS_2", "DS_3"]
        elif all(column in dmg_result.columns for column in ["insignific", "moderate", "heavy", "complete"]):
            dmg_states = ["insignific", "moderate", "heavy", "complete"]
        elif all(column in dmg_result.columns for column in ["ds-none", "ds-slight", "ds-moderat", "ds-extensi",
                                                             "ds-complet"]):
            dmg_states = ["ds-none", "ds-slight", "ds-moderat", "ds-extensi", "ds-complet"]
        else:
            raise ValueError("Invalid damage state names. Cannot create mapped max damage state.")

        guids = dmg_result[["guid"]]
        max_val = dmg_result[dmg_states].max(axis=1)
        max_key = dmg_result[dmg_states].idxmax(axis=1)
        dmg_concat = pd.concat([guids, max_val, max_key], axis=1)
        dmg_concat.rename(columns={0: "max_prob", 1: "max_state"}, inplace=True)

        return dmg_concat



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np


class JoplinEmpirRestorUtil:
    """Utility methods for the Joplin restoration analysis."""

    # Empirical coefficients mean and sigma for lognormal distribution
    FL_COEF = np.zeros((4, 4, 2), dtype=float)
    # initial functionality is FL1
    # target functionality is FL0
    FL_COEF[0, 0, :] = [5.90, 0.5]
    # other (for consistency)
    FL_COEF[1, 1, :] = [0.0, 1.0]
    FL_COEF[1, 2, :] = [0.0, 1.0]
    FL_COEF[1, 3, :] = [0.0, 1.0]

    # initial functionality is FL2
    # target functionality is FL0
    FL_COEF[1, 0, :] = [6.06, 0.32]
    # target functionality is FL1
    FL_COEF[1, 1, :] = [5.87, 0.35]
    # other (for consistency)
    FL_COEF[1, 2, :] = [0.0, 1.0]
    FL_COEF[1, 3, :] = [0.0, 1.0]

    # initial functionality is FL3
    # target functionality is FL0
    FL_COEF[2, 0, :] = [5.90, 0.60]
    # target functionality is FL1
    FL_COEF[2, 1, :] = [5.74, 0.55]
    # target functionality is FL2
    FL_COEF[2, 2, :] = [5.56, 0.46]
    # other (for consistency)
    FL_COEF[2, 3, :] = [0.0, 1.0]

    # initial functionality is FL4
    # target functionality is FL0
    FL_COEF[3, 0, :] = [6.60, 0.53]
    # target functionality is FL1
    FL_COEF[3, 1, :] = [6.39, 0.48]
    # target functionality is FL2
    FL_COEF[3, 2, :] = [6.49, 0.6]
    # target functionality is FL3
    FL_COEF[3, 3, :] = [6.13, 0.33]



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import json
from pyincore.models.mapping import Mapping


class MappingSet:
    """class for dfr3 mapping.

    Args:
        metadata (dict): mapping metadata.

    """

    def __init__(self, metadata):
        self.id = metadata["id"]
        self.name = metadata["name"]
        self.hazard_type = metadata["hazardType"]
        self.inventory_type = metadata['inventoryType']

        if 'dataType' in metadata:
            self.data_type = metadata["dataType"]
        else:
            self.data_type = "incore:dfr3MappingSet"

        mappings = []
        for m in metadata['mappings']:
            if isinstance(m, Mapping):
                mappings.append(m)
            else:
                # enforce convert dictionary to mapping entry object
                mappings.append(Mapping(m['entry'], m['rules']))
        self.mappings = mappings

        self.mapping_type = metadata['mappingType']

    @classmethod
    def from_json_str(cls, json_str):
        """Get dfr3 mapping from json string.

        Args:
            json_str (str): JSON of the Dataset.

        Returns:
            obj: dfr3 mapping from JSON.

        """

        return cls(json.loads(json_str))

    @classmethod
    def from_json_file(cls, file_path, data_type):
        """Get dfr3 mapping from the file.

        Args:
            file_path (str): json file path that holds the definition of a dfr3 curve.
            data_type (str): mapping dataset type

        Returns:
            obj: dfr3 mapping from file.

        """
        with open(file_path, "r") as f:
            metadata = json.load(f)
            metadata["dataType"] = data_type

            instance = cls(metadata)

        return instance



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

from pyincore.analyses.indp.infrastructurenode import InfrastructureNode
from pyincore.analyses.indp.infrastructurearc import InfrastructureArc
from pyincore.analyses.indp.infrastructureinterdeparc import InfrastructureInterdepArc
from pyincore.analyses.indp.infrastructurenetwork import InfrastructureNetwork


class InfrastructureUtil:

    @staticmethod
    def load_infrastructure_array_format_extended(power_nodes, power_arcs, water_nodes, water_arcs,
                                                  interdep, cost_scale=1.0,
                                                  extra_commodity=None):
        """
        This function reads the infrastructure network from file in the extended format

        Args:
            power_nodes:
            power_arcs:
            water_nodes:
            water_arcs:
            interdep:
            cost_scale (float): The factor by which all cost values has to multiplied. The default is 1.0.
            extra_commodity (dict): ( only for extended format of input data) List of extra-commodities in the
            analysis. The default is None, which only considers a main commodity.

        Returns:
            G (class:`~infrastructure.InfrastructureNetwork`): The object containing the network data.

        """
        net_names = {'Water': 1, 'Gas': 2, 'Power': 3, 'Telecommunication': 4}  # !!!
        G = InfrastructureNetwork("Test")
        global_index = 0

        for index, data in enumerate([power_nodes, water_nodes]):
            if index == 0:
                net = net_names["Power"]
            else:
                net = net_names["Water"]
            for v in data.iterrows():
                try:
                    node_id = v[1]['ID']
                except KeyError:
                    node_id = v[1]['nodenwid']
                n = InfrastructureNode(global_index, net, int(node_id))
                G.G.add_node((n.local_id, n.net_id), data={'inf_data': n})
                global_index += 1
                node_main_data = G.G.nodes[(n.local_id, n.net_id)]['data']['inf_data']
                node_main_data.reconstruction_cost = float(v[1]['q_ds_3']) * cost_scale
                node_main_data.oversupply_penalty = float(v[1]['Mp']) * cost_scale
                node_main_data.undersupply_penalty = float(v[1]['Mm']) * cost_scale
                node_main_data.demand = float(v[1]['Demand'])
                if 'guid' in v[1].index.values:
                    node_main_data.guid = v[1]['guid']
                resource_names = [x for x in list(v[1].index.values) if x[:2] == 'p_']
                if len(resource_names) > 0:
                    n.set_resource_usage(resource_names)
                    for rc in resource_names:
                        n.resource_usage[rc] = v[1][rc]
                else:
                    n.resource_usage['p_'] = 1
                if extra_commodity:
                    n.set_extra_commodity(extra_commodity[net])
                    for layer in extra_commodity[net]:
                        ext_com_data = G.G.nodes[(n.local_id, n.net_id)]['data']['inf_data'].extra_com[layer]
                        ext_com_data['oversupply_penalty'] = float(v[1]['Mp_' + layer]) * cost_scale
                        ext_com_data['undersupply_penalty'] = float(v[1]['Mm_' + layer]) * cost_scale
                        ext_com_data['demand'] = float(v[1]['Demand_' + layer])

        for index, data in enumerate([power_arcs, water_arcs]):
            if index == 0:
                net = net_names["Power"]
            else:
                net = net_names["Water"]
            for v in data.iterrows():
                try:
                    start_id = v[1]['Start Node']
                    end_id = v[1]['End Node']
                except KeyError:
                    start_id = v[1]['fromnode']
                    end_id = v[1]['tonode']
                for duplicate in range(2):
                    if duplicate == 0:
                        a = InfrastructureArc(int(start_id), int(end_id), net)
                    elif duplicate == 1:
                        a = InfrastructureArc(int(end_id), int(start_id), net)
                    G.G.add_edge((a.source, a.layer), (a.dest, a.layer), data={'inf_data': a})
                    arc_main_data = G.G[(a.source, a.layer)][(a.dest, a.layer)]['data']['inf_data']
                    arc_main_data.flow_cost = float(v[1]['c']) * cost_scale
                    arc_main_data.reconstruction_cost = float(v[1]['f']) * cost_scale
                    arc_main_data.capacity = float(v[1]['u'])
                    if 'guid' in v[1].index.values:
                        arc_main_data.guid = v[1]['guid']
                    resource_names = [x for x in list(v[1].index.values) if x[:2] == 'h_']
                    if len(resource_names) > 0:
                        a.set_resource_usage(resource_names)
                        for rc in resource_names:
                            a.resource_usage[rc] = v[1][rc]
                    else:
                        a.resource_usage['h_'] = 1
                    if extra_commodity:
                        a.set_extra_commodity(extra_commodity[net])
                        for layer in extra_commodity[net]:
                            ext_com_data = \
                                G.G[(a.source, a.layer)][(a.dest, a.layer)]['data']['inf_data'].extra_com[layer]
                            ext_com_data['flow_cost'] = float(v[1]['c_' + layer]) * cost_scale

        for v in interdep.iterrows():
            if v[1]['Type'] == 'Physical':
                i = int(v[1]['Dependee Node'])
                net_i = net_names[v[1]['Dependee Network']]
                j = int(v[1]['Depender Node'])
                net_j = net_names[v[1]['Depender Network']]
                a = InfrastructureInterdepArc(i, j, net_i, net_j, gamma=1.0)
                G.G.add_edge((a.source, a.source_layer), (a.dest, a.dest_layer), data={'inf_data': a})
                if extra_commodity:
                    a.set_extra_commodity(extra_commodity[net_i])
                    a.set_extra_commodity(extra_commodity[net_j])

        return G

    @staticmethod
    def add_from_csv_failure_scenario(G, sample, initial_node, initial_link):
        """
        This function reads initial damage data from file in the from_csv format, and applies it to the infrastructure
        network. This format only considers one magnitude value (0), and there can be as many samples from that magnitude.

        Parameters
        ----------
        G : :class:`~infrastructure.InfrastructureNetwork`
             The object containing the network data.
        sample : int
            Sample number of the initial damage scenario,

        Returns
        -------
        None.

        """
        for index, row in initial_node.iterrows():
            raw_n = row["name"].split(",")
            n = (int(raw_n[0].strip(' )(')), int(raw_n[1].strip(' )(')))
            state = float(row[sample + 1])
            G.G.nodes[n]['data']['inf_data'].functionality = state
            G.G.nodes[n]['data']['inf_data'].repaired = state

        for index, row in initial_link.iterrows():
            raw_uv = row["name"].split(",")
            u = (int(raw_uv[0].strip(' )(')), int(raw_uv[1].strip(' )(')))
            v = (int(raw_uv[2].strip(' )(')), int(raw_uv[3].strip(' )(')))
            state = float(row[sample + 1])
            if state == 0.0:
                G.G[u][v]['data']['inf_data'].functionality = state
                G.G[u][v]['data']['inf_data'].repaired = state

                G.G[v][u]['data']['inf_data'].functionality = state
                G.G[v][u]['data']['inf_data'].repaired = state



from __future__ import division
import csv
import pandas as pd
import networkx as nx
import copy
import math
from pyincore import GeoUtil, InventoryDataset


class TransportationRecoveryUtil:

    @staticmethod
    def NBI_coordinate_mapping(NBI_file):
        """Coordinate in NBI is in format of xx(degree)xx(minutes)xx.xx(seconds)
        map it to traditional xx.xxxx in order to create shapefile.

        Args:
            NBI_file (str): Filename of a NBI file.

        Returns:
            dict: NBI.

        """
        NBI = pd.read_csv(NBI_file)
        NBI['LONG_017'] = NBI['LONG_017'].apply(lambda x: -1 * (GeoUtil.degree_to_decimal(x)))
        NBI['LAT_016'] = NBI['LAT_016'].apply(lambda x: GeoUtil.degree_to_decimal(x))

        return NBI

    @staticmethod
    def get_average_daily_traffic(bridges, NBI_shapefile):
        NBI = InventoryDataset(NBI_shapefile)
        NBI_features = list(NBI.inventory_set)

        ADT = {}
        for bridge in bridges:
            # convert lon and lat to the right format
            bridge_coord = GeoUtil.get_location(bridge)
            nearest_feature, distance = GeoUtil.find_nearest_feature(NBI_features, bridge_coord)

            ADT[bridge['properties']['guid']] = nearest_feature['properties']['ADT_029']

        return ADT

    @staticmethod
    def convert_dmg_prob2state(dmg_results_filename):
        """Upstream bridge damage analysis will generate a dmg result file with the probability
        of each damage state; here determine what state using the maximum probability.

        Args:
            dmg_results_filename (str): Filename of a damage results file.

        Returns:
            dict: Bridge damage values.
            list: Unrepaired bridge.

        """
        bridge_damage_value = {}
        unrepaired_bridge = []

        with open(dmg_results_filename, 'r') as f:
            reader = csv.reader(f)
            next(reader)
            for row in reader:
                state_id = row[0]
                mean_damage = float(row[10])

                if mean_damage > 0 and mean_damage < 0.25:
                    bridge_damage_value[state_id] = 1
                elif mean_damage >= 0.25 and mean_damage < 0.5:
                    bridge_damage_value[state_id] = 2
                elif mean_damage >= 0.5 and mean_damage < 0.75:
                    bridge_damage_value[state_id] = 3
                elif mean_damage >= 0.75 and mean_damage <= 1:
                    bridge_damage_value[state_id] = 4
                else:
                    raise ValueError('mean damage should not larger than 1!')

            unrepaired_bridge = list(bridge_damage_value.keys())

        return bridge_damage_value, unrepaired_bridge

    @staticmethod
    def nw_reconstruct(node_df, arc_df, adt_data):
        """

        Args:
            node_df (pd.DataFrame): A node in _node_.csv.
            arc_df (pd.DataFrame): A node in edge in _edge_.csv.
            adt_data (pd.DataFrame): Average daily traffic flow.

        Returns:
            obj: Network

        """
        # create network
        network = nx.Graph()

        # add nodes to the network
        network.add_nodes_from(node_df['guid'])

        # add arcs to the network
        for i in range(len(arc_df)):
            fromnode = \
                node_df.loc[node_df['ID'] == arc_df['fromnode'][i], 'guid'].values[
                    0]
            tonode = \
                node_df.loc[node_df['ID'] == arc_df['tonode'][i], 'guid'].values[0]
            dis = arc_df['len_mile'][i] / arc_df['freeflowsp'][i]
            network.add_edge(fromnode, tonode, distance=dis, adt=adt_data[arc_df['guid'][i]])

        return network

    @staticmethod
    def traveltime_freeflow(temp_network):
        """A travel time calculation.

        Args:
            temp_network (obj): The investigated network.

        Returns:
            float: Travel efficiency.

        """
        network = copy.deepcopy(temp_network)

        for Ed in temp_network.edges():
            if network.edges[Ed[0], Ed[1]]['Damage_Status'] > 2:
                network.remove_edge(Ed[0], Ed[1])
            elif network.edges[Ed[0], Ed[1]]['Damage_Status'] == 2:
                network.edges[Ed[0], Ed[1]]['distance'] \
                    = network.edges[Ed[0], Ed[1]]['distance'] / 0.5
            elif network.edges[Ed[0], Ed[1]]['Damage_Status'] == 1:
                network.edges[Ed[0], Ed[1]]['distance'] \
                    = network.edges[Ed[0], Ed[1]]['distance'] / 0.75

        num_node = len(network.nodes())
        distance = [[0 for x in range(num_node)] for y in range(num_node)]

        tdistance = dict(nx.all_pairs_dijkstra_path_length(network,
                                                           weight='distance'))
        i = 0
        for key1, value1 in tdistance.items():
            j = 0
            for key2 in list(value1.keys()):
                distance[i][j] = tdistance[key1][key2]
                j += 1
            i += 1

        travel_efficiency = 0
        for i in range(num_node):
            for j in range(num_node):
                if i != j:
                    if distance[i][j] == 0:
                        distance[i][j] = math.inf
                    travel_efficiency += 1 / distance[i][j]

        return travel_efficiency



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import pandas as pd


def gams_to_dataframes(iNum, vars, H, L, soln):
    # CG0
    CG0 = vars.get('CG', x=soln[0])

    # CH0
    CH0 = vars.get('CH', x=soln[0])

    # CMI0
    CMI0 = vars.get('CMI', x=soln[0])

    # CMO0
    CMO0 = vars.get('CMO', x=soln[0])

    # CN0
    CN0 = vars.get('CN', x=soln[0])

    # CPI0
    CPI0 = vars.get('CPI', x=soln[0])

    # CX0
    CX0 = vars.get('CX', x=soln[0])

    # D0
    D0 = vars.get('D', x=soln[0])

    # DD0
    DD0 = vars.get('DD', x=soln[0])

    # DS0
    DS0 = vars.get('DS', x=soln[0])

    # FD
    FD0 = vars.get('FD', x=soln[0])

    # IGT
    IGT0 = vars.get('IGT', x=soln[0])

    # KS
    KS0 = vars.get('KS', x=soln[0])

    # LAS
    # LAS0 = vars.get('LAS', x=soln[0])

    # HH
    HH0 = vars.get('HH', x=soln[0])

    # HN
    HN0 = vars.get('HN', x=soln[0])

    # HW
    HW0 = vars.get('HW', x=soln[0])

    # M
    M0 = vars.get('M', x=soln[0])

    # N
    N0 = vars.get('N', x=soln[0])

    # NKI
    NKI0 = vars.get('NKI', x=soln[0])

    # LNFOR
    # LNFOR0 = vars.get('LNFOR', x=soln[0])

    # KPFOR
    KPFOR0 = vars.get('KPFOR', x=soln[0])

    # GVFOR
    GVFOR0 = vars.get('GVFOR', x=soln[0])

    # P
    P0 = vars.get('P', x=soln[0])

    # PD
    PD0 = vars.get('PD', x=soln[0])

    # PVA
    PVA0 = vars.get('PVA', x=soln[0])

    # RA
    RA0 = vars.get('RA', x=soln[0])

    # R
    R0 = vars.get('R', x=soln[0])

    # S
    S0 = vars.get('S', x=soln[0])

    # SPI
    SPI0 = vars.get('SPI', x=soln[0])

    # V
    V0 = vars.get('V', x=soln[0])

    # Y
    Y0 = vars.get('Y', x=soln[0])

    # Yd
    YD0 = vars.get('YD', x=soln[0])

    for i in range(iNum):
        CGL = vars.get('CG', x=soln[i + 1])
        CHL = vars.get('CH', x=soln[i + 1])
        CMIL = vars.get('CMI', x=soln[i + 1])
        CMOL = vars.get('CMO', x=soln[i + 1])
        CNL = vars.get('CN', x=soln[i + 1])
        CPIL = vars.get('CPI', x=soln[i + 1])
        CXL = vars.get('CX', x=soln[i + 1])
        DL = vars.get('D', x=soln[i + 1])
        DDL = vars.get('DD', x=soln[i + 1])
        DSL = vars.get('DS', x=soln[i + 1])
        FDL = vars.get('FD', x=soln[i + 1])
        IGTL = vars.get('IGT', x=soln[i + 1])
        KSL = vars.get('KS', x=soln[i + 1])
        # LASL = vars.get('LAS', x=soln[i+1])
        HHL = vars.get('HH', x=soln[i + 1])
        HNL = vars.get('HN', x=soln[i + 1])
        HWL = vars.get('HW', x=soln[i + 1])
        ML = vars.get('M', x=soln[i + 1])
        NL = vars.get('N', x=soln[i + 1])
        NKIL = vars.get('NKI', x=soln[i + 1])
        # LNFORL = vars.get('LNFOR', x=soln[i+1])
        KPFORL = vars.get('KPFOR', x=soln[i + 1])
        GVFORL = vars.get('GVFOR', x=soln[i + 1])
        PL = vars.get('P', x=soln[i + 1])
        PDL = vars.get('PD', x=soln[i + 1])
        PVAL = vars.get('PVA', x=soln[i + 1])
        RAL = vars.get('RA', x=soln[i + 1])
        RL = vars.get('R', x=soln[i + 1])
        SL = vars.get('S', x=soln[i + 1])
        SPIL = vars.get('SPI', x=soln[i + 1])
        VL = vars.get('V', x=soln[i + 1])
        YL = vars.get('Y', x=soln[i + 1])
        YDL = vars.get('YD', x=soln[i + 1])

        # DFCG.L(I,G)      = CG.L(I,G)-CG0(I,G);
        DFCG = CGL - CG0

        # DFFD.L(F,Z)      = FD.L(F,Z)-FD0(F,Z);
        DFFD = FDL - FD0

        # DK.L(K,Z)        = FD.L(K,Z)-FD0(K,Z);
        DK = KSL - KS0

        # DY.L(Z)          = Y.L(Z)-Y0(Z);
        # DY = YL - Y0
        DY = (YL / CPIL) - Y0

        # DDS.L(I)         = DS.L(I)-DS0(I);
        DDS = DSL - DS0

        # DDD.L(I)         = DD.L(I) - DD0(I);
        DDD = DDL - DD0

        # DCX.L(I)         = CX.L(I) -CX0(I);
        DCX = CXL - CX0

        # DCH.L(I,H)       = CH.L(I,H) - CH0(I,H);
        DCH = CHL - CH0

        # DRR.L(F,Z)       = R.L(F,Z)-R0(F,Z);
        DR = RL - R0

        # DCMI.L(L)        = CMI.L(L) - CMI0(L);
        DCMI = CMIL - CMI0

        # DCMO.L(CM)       = CMO.L(CM) - CMO0(CM);
        DCMO = CMOL - CMO0

        DM = ML - M0

        DV = VL - V0

        DN = NL - N0

        s_name = 'Simulation ' + str(i + 1)

    '''
    export domestic supply, household income (gross income), number of household, and factor demand
    '''
    # domestic supply
    cols = {'DS0': DS0,
            'DSL': DSL}
    domestic_supply_df = pd.DataFrame.from_dict(cols)
    domestic_supply_df.index.name = 'Sectors'
    # df.to_csv('domestic-supply.csv')

    # gross income
    cols = {'Y0': Y0.loc[H],
            'YL': YL.loc[H]}
    gross_income_df = pd.DataFrame.from_dict(cols)
    gross_income_df.index.name = 'Household Group'
    # df.to_csv('gross-income.csv')

    # household count
    cols = {'HH0': HH0,
            'HHL': HHL}
    household_count_df = pd.DataFrame.from_dict(cols)
    household_count_df.index.name = 'Household Group'
    # df.to_csv('household-count.csv')

    # pre-disaster-factor-demand
    pre_disaster_demand_df = FD0.loc[L]
    pre_disaster_demand_df.index.name = 'Labor Group'
    # df.to_csv('pre-disaster-factor-demand.csv')

    # post-disaster-factor-demand
    post_disaster_demand_df = FDL.loc[L]
    post_disaster_demand_df.index.name = 'Labor Group'
    # df.to_csv('post-disaster-factor-demand.csv')

    return domestic_supply_df, gross_income_df, household_count_df, pre_disaster_demand_df, post_disaster_demand_df



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class EpfUtil:
    """Utility methods for the electric power facility damage analysis."""
    EPF_HAZUS_FRAGILITY_KEYS = {
        "ESS1", "Low Voltage (115 KV) Substation (Anchored/Seismic Components)",
        "ESS2", "Low Voltage (115 KV) Substation (Unanchored/Standard Components)",
        "ESS3", "Medium Voltage (230 KV) Substation (Anchored/Seismic Components)",
        "ESS4", "Medium Voltage (230 KV) Substation (Unanchored/Standard Components)",
        "ESS5", "High Voltage (500 KV) Substation (Anchored/Seismic Components)",
        "ESS6", "High Voltage (500 KV) Substation (Unanchored/Standard Components)",
        "EPP1", "Small Generation Facility (Anchored Components)",
        "EPP2", "Small Generation Facility (Unanchored Components)",
        "EPP3", "Medium/Large Generation Facility (Anchored Components)",
        "EPP4", "Medium/Large Generation Facility (Unanchored Components)",
        "EDC1", "Distribution Circuit (Seismic Components)",
        "EDC2", "Distribution Circuit (Standard Components)"
    }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import fiona
import csv
import copy
import uuid
import networkx as nx
import matplotlib.pyplot as plt

from shapely.geometry import shape, LineString, Point, mapping
from fiona.crs import from_epsg


class NetworkUtil:
    @staticmethod
    def build_link_by_node(node_filename, graph_filename, id_field, out_filename):
        """Create line dataset based on node shapefile and graph file graph should be in csv format.

        Args:
            node_filename (string):  A node shapefile file name pull path with *.shp file extension.
            graph_filename (string): A graph csv file name pull path.
            id_field (string): A field name for node shapefiles unique id that matches the information in the graph.
            out_filename (string): A line file name pull path that will be newly created by the process.

        Returns:
            bool: To indicated if the line was created or not.

        """
        # read graph
        with open(graph_filename) as f:
            reader = csv.reader(f, delimiter=',')
            graph_list = list(reader)
        # remove the first element, which is a header, from a list
        graph_list.pop(0)

        # read node shapefile and make it as a list so it can be accessed for the building line file
        # also, add guid in the field
        innode = fiona.open(node_filename)

        # create list of each shapefile entry
        node_list = []
        node_id_list = []

        # create a schema for output line file
        schema = {
            'geometry': 'LineString',
            'properties': {
                'linkid': 'str:10',
                'guid': 'str:30',
                'from_node': 'str:10',
                'to_node': 'str:10'
            }
        }

        for in_feature in innode:
            # build shape feature
            tmp_feature = copy.deepcopy(in_feature)
            tmp_feature['properties']['guid'] = str(uuid.uuid4())
            node_id = str(tmp_feature['properties'][id_field])
            node_list.append(tmp_feature)
            node_id_list.append(node_id)

        # iterate graph and find from-node and to-node geography
        line_geom_list = []
        line_id_list = []
        line_from_list = []
        line_to_list = []
        for graph_row in graph_list:
            line_id = str(graph_row[0])
            from_id = str(graph_row[1])
            to_id = str(graph_row[2])
            line_id_list.append(line_id)
            line_from_list.append(from_id)
            line_to_list.append(to_id)

            # initialize the location
            from_location_in_node_list = 0
            to_location_in_node_list = 0
            for i in range(len(node_id_list)):
                if str(node_id_list[i]) == from_id:
                    from_location_in_node_list = i
                if str(node_id_list[i]) == to_id:
                    to_location_in_node_list = i

            from_geo = node_list[from_location_in_node_list]['geometry']
            to_geo = node_list[to_location_in_node_list]['geometry']
            line_geom = LineString([Point(shape(from_geo).coords), Point(shape(to_geo).coords)])
            line_geom_list.append(line_geom)

        # create line feature
        with fiona.open(out_filename, 'w', crs=from_epsg(4326), driver='ESRI Shapefile', schema=schema) as layer:
            for i in range(len(line_geom_list)):
                # filling schema
                schema['geometry'] = mapping(line_geom_list[i])
                schema['properties']['linkid'] = line_id_list[i]
                schema['properties']['guid'] = str(uuid.uuid4())
                schema['properties']['from_node'] = line_from_list[i]
                schema['properties']['to_node'] = line_to_list[i]

                layer.write(schema)

        return True

    @staticmethod
    def build_node_by_link(link_filename, link_id_field, fromnode_field, tonode_field, out_node_filename,
                           out_graph_filename):
        """Create node dataset based on line shapefile and graph file graph should be in csv format

        Args:
            link_filename (string):  line shapefile file name pull path with *.shp file extension.
            link_id_field (string): line shapefile unique id field
            fromnode_field (string): field name for fromnode in line shapefile
            tonode_field (string): field name for tonode in line shapefile
            out_node_filename (string): output node shapefile name with *.shp extension
            out_graph_filename (string): output graph csv file name with *.csv extension

        Returns:
            bool: To indicated if the shapefile and graph were created or not.

        """
        # read line shapefile
        linefile = fiona.open(link_filename)

        node_list = []
        node_id_list = []
        graph_list = []

        for line in linefile:
            line_geom = shape(line['geometry'])
            seg_coord_list = list(line_geom.coords)

            # to check if this is a multiline string
            if len(seg_coord_list) > 2:
                print("The line shapefile is a multiline string. The process will be aborted")
                return False

            line_id = str(line['properties'][link_id_field])
            fromnode_id = str(line['properties'][fromnode_field])
            tonode_id = str(line['properties'][tonode_field])
            fromnode_coord = seg_coord_list[0]
            tonode_coord = seg_coord_list[1]
            graph_line_list = []
            graph_line_list.append(line_id)
            graph_line_list.append(fromnode_id)
            graph_line_list.append(tonode_id)
            graph_list.append(graph_line_list)

            # check if to from node is in the node_id_list
            is_new_node = True
            for i in range(len(node_id_list)):
                if fromnode_id == node_id_list[i]:
                    is_new_node = False
                    break
            if is_new_node:
                node_id_list.append(fromnode_id)
                node_list.append(Point(fromnode_coord))

            # check if to to node is in the node_id_list
            is_new_node = True
            for i in range(len(node_id_list)):
                if tonode_id == node_id_list[i]:
                    is_new_node = False
                    break
            if is_new_node:
                node_id_list.append(tonode_id)
                node_list.append(Point(tonode_coord))

        # create a schema for output line file
        schema = {
            'geometry': 'Point',
            'properties': {
                'nodeid': 'str:10',
                'guid': 'str:30'
            }
        }

        # create output node feature
        with fiona.open(out_node_filename, 'w', crs=from_epsg(4326), driver='ESRI Shapefile', schema=schema) as layer:
            for i in range(len(node_id_list)):
                # filling schema
                schema['geometry'] = mapping(node_list[i])
                schema['properties']['nodeid'] = node_id_list[i]
                schema['properties']['guid'] = str(uuid.uuid4())

                layer.write(schema)

        with open(out_graph_filename, "w", newline='') as f:
            writer = csv.writer(f)
            writer.writerows([[link_id_field, fromnode_field, tonode_field]])
            writer.writerows(graph_list)

        return True

    @staticmethod
    def create_network_graph_from_link(link_file, fromnode_fldname, tonode_fldname, is_directed=False):
        """Create network graph from field.

        Args:
            link_file (str):  A name of a geo dataset resource recognized by Fiona package.
            fromnode_fldname (str): Line feature, from node field name.
            tonode_fldname (str): Line feature, to node field name.
            is_directed (bool, optional (Defaults to False)): Graph type. True for directed Graph,
                False for Graph.

        Returns:
            obj: A graph from field.
            dict: Coordinates.

        """
        # iterate link
        fromnode_list = []
        tonode_list = []
        node_list = []
        size = 0

        indataset = fiona.open(link_file)

        for line_feature in indataset:
            from_node_val = None
            if fromnode_fldname in line_feature["properties"]:
                from_node_val = line_feature['properties'][fromnode_fldname]
            elif fromnode_fldname.lower() in line_feature["properties"]:
                from_node_val = line_feature['properties'][fromnode_fldname.lower()]
            to_node_val = None
            if tonode_fldname in line_feature["properties"]:
                to_node_val = line_feature['properties'][tonode_fldname]
            elif tonode_fldname.lower() in line_feature["properties"]:
                to_node_val = line_feature['properties'][tonode_fldname.lower()]
            fromnode_list.append(from_node_val - 1)
            tonode_list.append(to_node_val - 1)
            node_list.append(from_node_val - 1)
            node_list.append(to_node_val - 1)

        # create node list
        node_list.sort()
        node_list = list(set(node_list))

        # create graph
        graph = None
        if is_directed:
            graph = nx.DiGraph()
        else:
            graph = nx.Graph()

        for i in range(len(fromnode_list)):
            graph.add_edge(fromnode_list[i], tonode_list[i])

        # initialize coords dictionary
        coords = dict((i, None) for i in range(len(node_list) - 1))

        # create coordinates
        node_coords_list = [None] * (len(node_list))
        for line_feature in indataset:
            from_node_val = None
            if fromnode_fldname in line_feature["properties"]:
                from_node_val = line_feature['properties'][fromnode_fldname]
            elif fromnode_fldname.lower() in line_feature["properties"]:
                from_node_val = line_feature['properties'][fromnode_fldname.lower()]
            to_node_val = None
            if tonode_fldname in line_feature["properties"]:
                to_node_val = line_feature['properties'][tonode_fldname]
            elif tonode_fldname.lower() in line_feature["properties"]:
                to_node_val = line_feature['properties'][tonode_fldname.lower()]
            line_geom = (line_feature['geometry'])
            coords_list = line_geom.get('coordinates')
            from_coord = coords_list[0]
            to_coord = coords_list[1]
            coords[int(from_node_val) - 1] = from_coord
            coords[int(to_node_val) - 1] = to_coord

        return graph, coords

    @staticmethod
    def plot_network_graph(graph, coords):
        """Plot graph.

        Args:
            graph (obj):  A nx graph to be drawn.
            coords (dict): Position coordinates.

        """
        # nx.draw(graph, coords, with_lables=True, font_weithg='bold')

        # other ways to draw
        nx.draw_networkx_nodes(graph, coords, cmap=plt.get_cmap('jet'), node_size=100, node_color='g')
        nx.draw_networkx_labels(graph, coords)
        nx.draw_networkx_edges(graph, coords, edge_color='r', arrows=True)
        plt.show()

    @staticmethod
    def read_network_graph_from_file(filename, is_directed=False):
        """Get network graph from filename.

        Args:
            filename (str):  A name of a geo dataset resource recognized by Fiona package.
            is_directed (bool, optional (Defaults to False)): Graph type. True for directed Graph,
                False for Graph.

        Returns:
            obj: A graph from field.
            dict: Coordinates.

        """
        geom = nx.read_shp(filename)
        node_coords = {k: v for k, v in enumerate(geom.nodes())}
        # create graph
        graph = None
        if is_directed:
            graph = nx.DiGraph()
        else:
            graph = nx.Graph()

        graph.add_nodes_from(node_coords.keys())
        l = [set(x) for x in geom.edges()]
        edg = [tuple(k for k, v in node_coords.items() if v in sl) for sl in l]

        graph.add_edges_from(edg)

        return graph, node_coords

    @staticmethod
    def validate_network_node_ids(network_dataset, fromnode_fldname, tonode_fldname, nodeid_fldname):
        """Check if the node id in from or to node exist in the real node id.

        Args:
            network_dataset (obj):  Network dataset
            fromnode_fldname (str): Line feature, from node field name.
            tonode_fldname (str): Line feature, to node field name.
            nodeid_fldname (str): Node field id name.

        Returns:
            bool: Validation of node existence.

        """
        # get link and node dataset
        link_dataset = network_dataset.links.get_inventory_reader()
        node_dataset = network_dataset.nodes.get_inventory_reader()
        validate = True
        # iterate link
        link_node_list = []
        for line_feature in link_dataset:
            from_node_val = None
            if fromnode_fldname in line_feature["properties"]:
                from_node_val = line_feature['properties'][fromnode_fldname]
            elif fromnode_fldname.lower() in line_feature["properties"]:
                from_node_val = line_feature['properties'][fromnode_fldname.lower()]
            to_node_val = None
            if tonode_fldname in line_feature["properties"]:
                to_node_val = line_feature['properties'][tonode_fldname]
            elif tonode_fldname.lower() in line_feature["properties"]:
                to_node_val = line_feature['properties'][tonode_fldname.lower()]
            link_node_list.append(from_node_val)
            link_node_list.append(to_node_val)

        # iterate node
        node_list = []
        for node_feature in node_dataset:
            node_val = None
            if nodeid_fldname in node_feature["properties"]:
                node_val = node_feature['properties'][nodeid_fldname]
            elif nodeid_fldname.lower() in node_feature["properties"]:
                node_val = node_feature['properties'][nodeid_fldname.lower()]
            node_list.append(node_val)

        link_node_list.sort()
        node_list.sort()
        link_node_list = list(set(link_node_list))
        node_list = list(set(node_list))

        for node in link_node_list:
            if node in node_list == False:
                validate = False

        return validate

    @staticmethod
    def merge_labeled_networks(graph_a, graph_b, edges_ab, directed=False):
        """Merges two networks, each distinguished by a label

        Args:
            graph_a (obj): labeled network a
            graph_b (obj): labeled network b
            edges_ab (pd.DataFrame): mapping containing links between network a and network b, column labels should
                                     correspond to the labels in each graph
            directed (bool): if the network is directed, use an additional column to determine edge direction

        Returns:

            obj: a new graph that integrates the two networks
        """

        # Define directionality when needed
        __left_to_right = 1
        __right_to_left = -1
        __no_direction = 0

        # Extract labels
        labels = list(edges_ab.columns)

        prefix_a = labels[0]
        prefix_b = labels[1]

        # Ensure data types are correct
        edges_ab[prefix_a] = edges_ab[prefix_a].astype('int64')
        edges_ab[prefix_b] = edges_ab[prefix_a].astype('int64')

        direction = None

        if directed:
            direction = labels[2]
            edges_ab[direction] = edges_ab[direction].astype('int64')

        # Merge the networks
        merged_graph = nx.union(graph_a, graph_b, rename=(prefix_a, prefix_b))

        # Use edge and direction to add connecting edges
        for idx, row in edges_ab.iterrows():
            if directed:
                if row[direction] == __left_to_right:
                    merged_graph.add_edge(f"{prefix_a}{row[prefix_a]}", f"{prefix_b}{row[prefix_b]}")
                elif row[direction] == __no_direction:
                    merged_graph.add_edge(f"{prefix_a}{row[prefix_a]}", f"{prefix_b}{row[prefix_b]}")
                    merged_graph.add_edge(f"{prefix_b}{row[prefix_b]}", f"{prefix_a}{row[prefix_a]}")
                else:
                    merged_graph.add_edge(f"{prefix_b}{row[prefix_b]}", f"{prefix_a}{row[prefix_a]}")
            else:
                merged_graph.add_edge(f"{prefix_a}{row[prefix_a]}", f"{prefix_b}{row[prefix_b]}")

        return merged_graph

    @staticmethod
    def extract_network_by_label(labeled_graph, prefix):
        """Given a network resulting from a labeled merging, extract only one of the networks based on its prefix

        Args:
            labeled_graph (obj): a graph obtained by labeling and merging two networks
            prefix (str): label of the network to extract

        Returns:

            obj: a new graph represented the network extracted using the label
        """

        # Filter the list of nodes based on prefix
        prefix_nodes = filter(lambda node_id: node_id.startswith(prefix), list(labeled_graph.nodes))

        # Extract the corresponding subgraph
        subgraph = labeled_graph.subgraph(prefix_nodes)

        # Construct the inverse mapping to get back to the original network
        de_prefixed = map(lambda node_id: int(node_id.lstrip(prefix)), prefix_nodes)
        de_mapping = dict(zip(prefix_nodes, de_prefixed))
        return nx.relabel_nodes(subgraph, de_mapping, copy=True)

    @staticmethod
    def create_network_graph_from_dataframes(df_nodes, df_links, sort='unsorted'):
        """Given a dataframe of nodes and a dataframe of links, assemble a network object.

        Args:
            df_nodes (pd.DataFrame):
            df_links (pd.DataFrame):
            sort:

        Returns:

        """
        graph = nx.DiGraph()  # Empty graph

        pos_x = df_nodes['geometry'].apply(lambda p: p.x).head()
        pos_y = df_nodes['geometry'].apply(lambda p: p.y).head()
        node_id = df_nodes['nodenwid']

        pos = {}
        pos_x = df_nodes['geometry'].apply(lambda p: p.x)
        
        pos_y = df_nodes['geometry'].apply(lambda p: p.y)
        for i, val in enumerate(df_nodes["nodenwid"]):
            pos[val] = (pos_x[i], pos_y[i])

        edges = [(x, y) for x, y in zip(df_links["fromnode"], df_links["tonode"])]
        edge = []

        if sort == 'sorted':
            for i, val in enumerate(df_links["linknwid"]):
                if df_links["direction"][i] == 1:
                    edge.append((df_links["fromnode"][i], df_links["tonode"][i]))
                else:
                    edge.append((df_links["tonode"][i], df_links["fromnode"][i]))
        elif sort == 'unsorted':
            for i, val in enumerate(df_links["linknwid"]):
                edge.append((df_links["fromnode"][i], df_links["tonode"][i]))

        graph.add_nodes_from(pos.keys())
        graph.add_edges_from(edge)

        for x, y, id in zip(pos_x, pos_y, node_id):
            graph.nodes[id]['pos'] = (x, y)

        for ii, node_id in enumerate(graph.nodes()):
            graph.nodes[node_id]["classification"] = df_nodes["utilfcltyc"][ii]

        return graph


# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


from pyincore.analyses.tornadoepndamage.tornadoepndamage import TornadoEpnDamage



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import errno
import os
import fiona


class NetworkData:
    """ Network data from Fiona package. Fiona can read and write data using GIS formats.

    Args:
        network_type (str): Network type.
        file_path (str): Path to a file with GIS layers.

    """

    def __init__(self, network_type: str, file_path: str):
        self.network_type = network_type
        if os.path.exists(file_path):
            self.file_path = file_path
        else:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), file_path)

    def get_inventory_reader(self):
        """ getter """
        filename = self.file_path
        if os.path.isdir(filename):
            layers = fiona.listlayers(filename)
            if len(layers) > 0:
                # for now, open the first shapefile
                return fiona.open(filename, layer=layers[0])
        else:
            return fiona.open(filename)



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import getpass
import hashlib
import json
import os
import shutil
import urllib.parse

import requests

from pyincore import globals as pyglobals
from pyincore.utils import return_http_response

logger = pyglobals.LOGGER


def update_hash_entry(mode, hashed_url=None, service_url=None):
    """
    Helper function to add/remove the hashed url to service.json file to keep track of
    the data coming from various repositories.
    modes = ["add", "edit", "clear"]
    Args:
        mode (str): String value indicating operation mode
        hashed_url (optional str): String value of the Hashed Service URL
        service_url (optional str): Actual Service URL value

    Returns: None

    """
    if mode not in ["clear", "edit", "add"]:
        logger.warning("Incorrect mode. Please check the mode.")
        return

    service_json = {}
    # to clear the entire cache data folder
    if mode == "clear":
        with open(pyglobals.PYINCORE_SERVICE_JSON, "w") as f:
            json.dump(service_json, f, indent=4)
        return

    # to add a hash entry
    if mode == "add" and (service_url is not None and hashed_url is not None):
        entry = {
                "service-name": "",
                "service-url": service_url,
                "hash": hashed_url,
                "description": ""
        }
        if not os.path.exists(pyglobals.PYINCORE_SERVICE_JSON):
            with open(pyglobals.PYINCORE_SERVICE_JSON, "w") as f:
                service_json[hashed_url] = entry
                json.dump(service_json, f, indent=4)
                return
    # read the current entries in service.json
    try:
        with open(pyglobals.PYINCORE_SERVICE_JSON, "r") as f:
            service_json = json.load(f)
    except FileNotFoundError:
        logger.warning("service.json file not present.")
        return

    if hashed_url is None:
        logger.warning("Need Hashed URL with edit mode")
        return

    # write back the data with or without a hash entry depending upon operation
    with open(pyglobals.PYINCORE_SERVICE_JSON, "w") as f:
        if hashed_url not in service_json and mode == "add":
            service_json[hashed_url] = entry
        elif hashed_url in service_json and mode == "edit":
            del service_json[hashed_url]

        json.dump(service_json, f, indent=4)
    return


class Client:
    """Incore service Client class. It handles connection to the server with INCORE services and user authentication."""

    def __init__(self):
        self.session = requests.session()

        # if .incore is not a directory, create it and add a cache directory
        if not os.path.isdir(pyglobals.PYINCORE_USER_CACHE):
            os.makedirs(pyglobals.PYINCORE_USER_CACHE)
            os.makedirs(pyglobals.PYINCORE_USER_DATA_CACHE)
            if not os.path.isdir(pyglobals.PYINCORE_USER_CACHE):
                logger.warn("Unable to create .incore directory.")
        # if .incore is a directory but there is no cache directory, create it
        if not os.path.isdir(pyglobals.PYINCORE_USER_DATA_CACHE):
            os.makedirs(pyglobals.PYINCORE_USER_DATA_CACHE)
            if not os.path.isdir(pyglobals.PYINCORE_USER_DATA_CACHE):
                logger.warn("Unable to create cache directory.")

    def get(self, url: str, params=None, timeout=(30, 600), **kwargs):
        """Get server connection response.

        Args:
            url (str): Service url.
            params (obj): Session parameters.
            timeout (tuple): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.get(url, params=params, timeout=timeout, **kwargs)
        return return_http_response(r)

    def post(self, url: str, data=None, json=None, timeout=(30, 600), **kwargs):
        """Post data on the server.

        Args:
            url (str): Service url.
            data (obj): Data to be posted on the server.
            json (obj): Description of the data, metadata json.
            timeout (tuple): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.post(url, data=data, json=json, timeout=timeout, **kwargs)
        return return_http_response(r)

    def put(self, url: str, data=None, timeout=(30, 600), **kwargs):
        """Put data on the server.

        Args:
            url (str): Service url.
            data (obj): Data to be put onn the server.
            timeout (tuple): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.put(url, data=data, timeout=timeout, **kwargs)
        return return_http_response(r)

    def delete(self, url: str, timeout=(30, 600), **kwargs):
        """Delete data on the server.

        Args:
            url (str): Service url.
            timeout (tuple): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.delete(url, timeout=timeout, **kwargs)
        return return_http_response(r)


class IncoreClient(Client):
    """IN-CORE service client class. It contains token and service root url."""

    def __init__(self, service_url: str = None, token_file_name: str = None, offline: bool = False):
        """

        Args:
            service_url (str): Service url.
            token_file_name (str): Path to file containing the authorization token.
            offline (bool): Flag to indicate offline mode or not.
        """
        super().__init__()
        self.offline = offline

        if not offline:
            if service_url is None or len(service_url.strip()) == 0:
                service_url = pyglobals.INCORE_API_PROD_URL
            self.service_url = service_url
            self.token_url = urllib.parse.urljoin(self.service_url, pyglobals.KEYCLOAK_AUTH_PATH)

            # hashlib requires bytes array for hash operations
            byte_url_string = str.encode(self.service_url)
            self.hashed_service_url = hashlib.sha256(byte_url_string).hexdigest()

            self.create_service_json_entry()

            # construct local directory and filename
            cache_data = pyglobals.PYINCORE_USER_DATA_CACHE
            if not os.path.exists(cache_data):
                os.makedirs(cache_data)

            self.hashed_svc_data_dir = os.path.join(cache_data, self.hashed_service_url)

            if not os.path.exists(self.hashed_svc_data_dir):
                os.makedirs(self.hashed_svc_data_dir)

            # store the token file in the respective repository's directory
            if token_file_name is None or len(token_file_name.strip()) == 0:
                token_file_name = "." + self.hashed_service_url + "_token"
            self.token_file = os.path.join(pyglobals.PYINCORE_USER_CACHE, token_file_name)

            authorization = self.retrieve_token_from_file()
            if authorization is not None:
                self.session.headers["Authorization"] = authorization
                print("Connection successful to IN-CORE services.", "pyIncore version detected:", pyglobals.PACKAGE_VERSION)

            else:
                if self.login():
                    print("Connection successful to IN-CORE services.", "pyIncore version detected:",
                          pyglobals.PACKAGE_VERSION)
        else:
            self.service_url = ""
            self.token_url = ""
            self.hashed_service_url = ""
            self.hashed_svc_data_dir = ""
            self.token_file = ""
            print("You are working with the offline version of IN-CORE.", "pyIncore version detected:", pyglobals.PACKAGE_VERSION)

    def login(self):
        for attempt in range(pyglobals.MAX_LOGIN_ATTEMPTS):
            try:
                username = input("Enter username: ")
                password = getpass.getpass("Enter password: ")
            except EOFError as e:
                logger.warning(e)
                raise e
            r = requests.post(self.token_url, data={'grant_type': 'password',
                                                    'client_id': pyglobals.CLIENT_ID,
                                                    'username': username, 'password': password})
            try:
                token = return_http_response(r).json()
                if token is None or token["access_token"] is None:
                    logger.warning("Authentication Failed.")
                    exit(0)
                authorization = str("bearer " + token["access_token"])
                self.store_authorization_in_file(authorization)
                self.session.headers['Authorization'] = authorization
                return True
            except Exception as e:
                logger.warning("Authentication failed, attempting login again.")
                print(e)


        logger.warning("Authentication failed.")
        exit(0)

    def store_authorization_in_file(self, authorization: str):
        """Store the access token in local file. If the file does not exist, this function creates it.

        Args:
            authorization (str): An authorization in the format "bearer access_token".

        """
        try:
            with open(self.token_file, 'w') as f:
                f.write(authorization)
        except IOError as e:
            logger.warning(e)

    def retrieve_token_from_file(self):
        """Attempts to retrieve authorization from a local file, if it exists.

        Returns:
            None if token file does not exist
            dict: Dictionary containing authorization in  the format "bearer access_token" if file exists, None otherwise

        """
        if not os.path.isfile(self.token_file):
            return None
        else:
            try:
                with open(self.token_file, 'r') as f:
                    auth = f.read().splitlines()
                    # check if token is valid
                    userinfo_url = urllib.parse.urljoin(self.service_url, pyglobals.KEYCLOAK_USERINFO_PATH)
                    r = requests.get(userinfo_url, headers={'Authorization': auth[0]})
                    if r.status_code != 200:
                        return None
                return auth[0]
            except IndexError:
                return None
            except OSError:
                return None

    def get(self, url: str, params=None, timeout=(30, 600), **kwargs):
        """Get server connection response.

        Args:
            url (str): Service url.
            params (obj): Session parameters.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.get(url, params=params, timeout=timeout, **kwargs)

        if r.status_code == 401:
            self.login()
            r = self.session.get(url, params=params, timeout=timeout, **kwargs)

        return return_http_response(r)

    def post(self, url: str, data=None, json=None, timeout=(30, 600), **kwargs):
        """Post data on the server.

        Args:
            url (str): Service url.
            data (obj): Data to be posted on the server.
            json (obj): Description of the data, metadata json.
            timeout (tuple[int,int]): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.post(url, data=data, json=json, timeout=timeout, **kwargs)

        if r.status_code == 401:
            self.login()
            r = self.session.post(url, data=data, json=json, timeout=timeout, **kwargs)

        return return_http_response(r)

    def put(self, url: str, data=None, timeout=(30, 600), **kwargs):
        """Put data on the server.

        Args:
            url (str): Service url.
            data (obj): Data to be put onn the server.
            timeout (tuple): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.put(url, data=data, timeout=timeout, **kwargs)

        if r.status_code == 401:
            self.login()
            r = self.session.put(url, data=data, timeout=timeout, **kwargs)

        return return_http_response(r)

    def delete(self, url: str, timeout=(30, 600), **kwargs):
        """Delete data on the server.

        Args:
            url (str): Service url.
            timeout (tuple): Session timeout.
            **kwargs: A dictionary of external parameters.

        Returns:
            obj: HTTP response.

        """
        r = self.session.delete(url, timeout=timeout, **kwargs)

        if r.status_code == 401:
            self.login()
            r = self.session.delete(url, timeout=timeout, **kwargs)

        return return_http_response(r)

    def create_service_json_entry(self):
        update_hash_entry("add", hashed_url=self.hashed_service_url, service_url=self.service_url)

    @staticmethod
    def clear_root_cache():
        # incase cache_data folder doesn't exist
        if not os.path.isdir(pyglobals.PYINCORE_USER_DATA_CACHE):
            logger.warning("User data cache does not exist")
            return None

        for root, dirs, files in os.walk(pyglobals.PYINCORE_USER_DATA_CACHE):
            for d in dirs:
                shutil.rmtree(os.path.join(root, d))
        update_hash_entry("clear")
        return None

    def clear_cache(self):
        """
            This function helps clear the data cache for a specific repository or the entire cache

            Returns: None

        """
        # incase cache_data folder doesn't exist
        if not os.path.isdir(pyglobals.PYINCORE_USER_DATA_CACHE):
            logger.warning("User data cache does not exist")
            return None

        if not os.path.isdir(self.hashed_svc_data_dir):
            logger.warning("Cached folder doesn't exist")
            return None

        shutil.rmtree(self.hashed_svc_data_dir)
        # clear entry from service.json
        update_hash_entry("edit", hashed_url=self.hashed_service_url)
        return


class InsecureIncoreClient(Client):
    """IN-CORE service client class that bypasses Ambassador auth. It contains token and service root url.

        Args:
            service_url (str): Service url.
            username (str): Username string.

        """

    def __init__(self, service_url: str = None, username: str = None):
        super().__init__()
        if service_url is None or len(service_url.strip()) == 0:
            service_url = pyglobals.INCORE_API_PROD_URL
        self.service_url = service_url
        if username is None or len(username.strip()) == 0:
            self.session.headers["x-auth-userinfo"] = pyglobals.INCORE_LDAP_TEST_USER_INFO
        else:
            user_info = "{\"preferred_username\": \"" + username + "\"}"
            self.session.headers["x-auth-userinfo"] = user_info



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd
from scipy.stats import lognorm
import time

from pyincore import BaseAnalysis, RepairService
from pyincore.analyses.buildingdamage.buildingutil import BuildingUtil


class CommercialBuildingRecovery(BaseAnalysis):
    """
    This analysis computes the recovery time needed for each commercial building from any damage states to receive the
    full restoration. Currently, supported hazards are tornadoes.

    The methodology incorporates the multi-layer Monte Carlo simulation approach and determines the two-step recovery
    time that includes delay and repair. The delay model was modified based on the REDi framework and calculated the
    end-result outcomes resulted from delay impeding factors such as post-disaster inspection, insurance claim,
    financing and government permit. The repair model followed the FEMA P-58 approach and was controlled by fragility
    functions.

    The outputs of this analysis is a CSV file with time-stepping recovery probabilities at the building level.

    Contributors
        | Science: Wanting Lisa Wang, John W. van de Lindt
        | Implementation: Wanting Lisa Wang, and NCSA IN-CORE Dev Team

    Related publications
        Wang, W.L., Watson, M., van de Lindt, J.W. and Xiao, Y., 2023. Commercial Building Recovery Methodology for Use
        in Community Resilience Modeling. Natural Hazards Review, 24(4), p.04023031.

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.repairsvc = RepairService(incore_client)

        super(CommercialBuildingRecovery, self).__init__(incore_client)

    def run(self):
        """Executes the commercial building recovery analysis.

        Returns:
            bool: True if successful, False otherwise.

        """
        # TODO: Start using seed
        seed = self.get_parameter("seed")

        num_samples = self.get_parameter("num_samples")
        result_name = self.get_parameter("result_name")

        repair_key = self.get_parameter("repair_key")
        if repair_key is None:
            repair_key = BuildingUtil.DEFAULT_REPAIR_KEY
            self.set_parameter("repair_key", repair_key)

        buildings = self.get_input_dataset("buildings").get_inventory_reader()
        buildings = list(buildings)
        sample_damage_states = self.get_input_dataset("sample_damage_states").get_dataframe_from_csv(low_memory=False)
        mcs_failure = self.get_input_dataset("mcs_failure").get_dataframe_from_csv(low_memory=False)
        redi_delay_factors = self.get_input_dataset("delay_factors").get_dataframe_from_csv(low_memory=False)
        building_dmg = self.get_input_dataset("building_dmg").get_dataframe_from_csv(low_memory=False)

        # Returns dataframe
        total_delay, recovery, time_stepping_recovery = self.commercial_recovery(buildings, sample_damage_states,
                                                                                 mcs_failure, redi_delay_factors,
                                                                                 building_dmg, num_samples)
        self.set_result_csv_data("total_delay", total_delay, result_name + "_delay", "dataframe")
        self.set_result_csv_data("recovery", recovery, result_name + "_recovery", "dataframe")
        self.set_result_csv_data("time_stepping_recovery", time_stepping_recovery,
                                 result_name + "_time_stepping_recovery", "dataframe")

        return True

    def commercial_recovery(self, buildings, sample_damage_states, mcs_failure, redi_delay_factors, building_dmg, num_samples):
        """
        Calculates commercial building recovery for buildings

        Args:
            buildings(list): Buildings dataset
            sample_damage_states (pd.DataFrame): Sample damage states
            redi_delay_factors (pd.DataFrame): Delay factors based on REDi framework
            mcs_failure (pd.DataFrame): Building inventory failure probabilities
            building_dmg (pd.DataFrame): Building damage states
            num_samples (int): number of sample scenarios to use

        Returns:
            dict: dictionary with id/guid and commercial recovery for each quarter

        """

        start_total_delay = time.process_time()
        total_delay = CommercialBuildingRecovery.total_delay(buildings, sample_damage_states, mcs_failure,
                                                             redi_delay_factors, building_dmg, num_samples)
        end_total_delay = time.process_time()
        print("Finished executing total_delay() in " + str(end_total_delay - start_total_delay) + " secs")

        recovery = self.recovery_rate(buildings, sample_damage_states, total_delay)
        end_recovery = time.process_time()
        print("Finished executing recovery_rate() in " + str(end_recovery - end_total_delay) + " secs")

        time_stepping_recovery = CommercialBuildingRecovery.time_stepping_recovery(recovery)
        end_time_stepping_recovery = time.process_time()
        print("Finished executing time_stepping_recovery() in " +
              str(end_time_stepping_recovery - end_recovery) + " secs")

        end_time = time.process_time()
        print("Analysis completed in " + str(end_time - start_total_delay) + " secs")

        return total_delay, recovery, time_stepping_recovery

    @staticmethod
    def total_delay(buildings, sample_damage_states, mcs_failure, redi_delay_factors, damage, num_samples):
        """ Calculates total delay by combining financial delay and other factors from REDi framework

        Args:
            buildings (list): List of buildings
            sample_damage_states (pd.DataFrame): Building inventory damage states.
            mcs_failure (pd.DataFrame): Building inventory failure probabilities
            redi_delay_factors (pd.DataFrame): Delay impeding factors such as post-disaster inspection, insurance claim,
                financing, and government permit based on building's damage state.
            damage (pd.DataFrame): Damage states for building structural damage
            num_samples (int): number of sample scenarios to use

        Returns:
            pd.DataFrame: Total delay time of all impeding factors from REDi framework.
        """

        # Obtain the commercial buildings in damage
        damage = mcs_failure[damage['haz_expose'] == 'yes']
        commercial = []
        commercial_archetypes = [6, 7, 8, 15, 16, 18, 19]
        for i, b in enumerate(buildings):
            if b['properties']['archetype'] in commercial_archetypes:
                commercial.append(b['properties']['guid'])
        commercial_pd = pd.DataFrame(commercial, columns=['guid'])
        commercial_damage = pd.merge(damage, commercial_pd, on='guid')

        # Obtain the column names
        colnames = [f'sample_{i}' for i in range(0, num_samples)]
        samples = np.zeros((len(commercial_damage), num_samples))
        delay_time = pd.DataFrame(samples, columns=colnames)
        delay_time.insert(0, 'guid', commercial_damage.reset_index(drop=True)['guid'])

        # Perform an inner join to ensure only buildings with damage states are processed
        merged_delay = pd.merge(sample_damage_states, delay_time, on='guid')

        # Obtain the guids
        merged_delay_guids = merged_delay['guid']

        # Obtain the damage states
        merged_delay_damage_states = merged_delay['sample_damage_states']

        # Convert to numpy
        samples_np = merged_delay.drop(columns=['guid', 'sample_damage_states']).to_numpy()

        # First, we decompose redi_delay_factors into two dictionaries that can be used to compute vector operations
        redi_idx = dict(zip(redi_delay_factors['Building_specific_conditions'], redi_delay_factors.index))

        # Next, we produce two intermediate numpy matrices: one for med and one for sdv
        redi_med = redi_delay_factors[['Ins_med', 'Enmo_med', 'Como_med', 'Per_med', 'Fin_med']].to_numpy()
        redi_sdv = redi_delay_factors[['Ins_sdv', 'Enmo_sdv', 'Como_sdv', 'Per_sdv', 'Fin_sdv']].to_numpy()

        # Define indices to facilitate interpretation of the code
        inspection_idx = 0
        engineer_idx = 1
        contractor_idx = 2
        permit_idx = 3
        financing_idx = 4

        for i, ds_list in enumerate(list(merged_delay_damage_states)):
            samples_mcs = ds_list.split(",")

            for j in range(num_samples):
                # TODO: ask why there are many more damage states than samples
                # Obtain the index for the corresponding damage state
                dmg_state_idx = redi_idx[samples_mcs[j]]

                # Use the index to select the appropriate mean and stdev vectors
                mean_vec = redi_med[dmg_state_idx, :]
                sdv_vec = redi_sdv[dmg_state_idx, :]

                # Compute the delay vector
                delay_vec = np.random.lognormal(np.log(mean_vec), sdv_vec)

                if dmg_state_idx == 3 or dmg_state_idx == 2:
                    delay_vec[4] = np.random.normal(mean_vec[4], sdv_vec[4])

                # Compute the delay using that vector, already computed in the prior step
                samples_np[i, j] = np.round(delay_vec[inspection_idx] +
                                            np.max([
                                                delay_vec[engineer_idx],
                                                delay_vec[financing_idx],
                                                delay_vec[contractor_idx]
                                            ]) +
                                            delay_vec[permit_idx])

        total_delay = pd.DataFrame(samples_np, columns=colnames)
        total_delay.insert(0, 'guid', merged_delay_guids)

        return total_delay

    def recovery_rate(self, buildings, sample_damage_states, total_delay):
        """ Gets total time required for each commercial building to receive full restoration. Determined by the
        combination of delay time and repair time

        Args:
            buildings (list): List of buildings
            sample_damage_states (pd.DataFrame): Samples' damage states
            total_delay (pd.DataFrame): Total delay time of financial delay and other factors from REDi framework.

        Returns:
            pd.DataFrame: Recovery time of all commercial buildings for each sample
        """

        repair_key = self.get_parameter("repair_key")
        repair_sets = self.repairsvc.match_inventory(self.get_input_dataset("dfr3_mapping_set"), buildings, repair_key)
        repair_sets_by_guid = {}  # get repair sets by guid so they can be mapped with output of monte carlo

        # This is sort of a workaround until we define Repair Curve models and abstract this out there
        for i, b in enumerate(buildings):
            # if building id has a matched repair curve set
            if b['id'] in repair_sets.keys():
                repair_sets_by_guid[b["properties"]['guid']] = repair_sets[b['id']]
            else:
                repair_sets_by_guid[b["properties"]['guid']] = None

        # Obtain the column names
        colnames = list(total_delay.columns)[1:]

        # Perform an inner join to ensure only buildings with damage states are processed
        merged_delay = pd.merge(total_delay, sample_damage_states, on='guid')

        # Obtain the guids
        merged_delay_guids = merged_delay['guid']

        # Obtain the damage states
        merged_delay_damage_states = merged_delay['sample_damage_states']

        # Convert to numpy
        samples_np = merged_delay.drop(columns=['guid', 'sample_damage_states']).to_numpy()
        num_samples = len(colnames)
        num_buildings = samples_np.shape[0]

        # Generate a long numpy matrix for combined N1, N2 samples
        samples_n1_n2 = np.zeros((num_buildings, num_samples * num_samples))

        # Now, we define an internal function to take care of the index for the prior case
        def idx(x, y):
            return x * num_samples + y

        for build in range(0, num_buildings):
            # Obtain the damage states
            mapped_repair = repair_sets_by_guid[merged_delay_guids.iloc[build]]
            samples_mcs = merged_delay_damage_states.iloc[build].split(",")

            # Use a lambda to obtain the damage state in numeric form. Note that since damage states are single digits,
            # it suffices to look at the last character and convert into an integer value. Do this computation once
            # per building only.
            extract_ds = lambda x: int(x[-1])
            samples_mcs_ds = list(map(extract_ds, samples_mcs))

            # Now, perform the two nested loops, using the indexing function to simplify the syntax.
            for i in range(0, num_samples):
                state = samples_mcs_ds[i]
                percent_func = np.random.random(num_samples)
                # NOTE: Even though the kwarg name is "repair_time", it actually takes  percent of functionality. DFR3
                # system currently doesn't have a way to represent the name correctly when calculating the inverse.
                if mapped_repair is not None:
                    repair_time = mapped_repair.repair_curves[state].solve_curve_for_inverse(
                        hazard_values={}, curve_parameters=mapped_repair.curve_parameters, **{"repair_time": percent_func}
                    ) / 7
                else:
                    repair_time = np.full(num_samples, np.nan)
                    
                for j in range(0, num_samples):
                    samples_n1_n2[build, idx(i, j)] = round(samples_np[build, i] + repair_time[j], 1)

        # Now, generate all the labels using list comprehension outside the loops
        colnames = [f'sample_{i}_{j}' for i in range(0, num_samples) for j in range(0, num_samples)]
        recovery_time = pd.DataFrame(samples_n1_n2, columns=colnames)
        recovery_time.insert(0, 'guid', merged_delay_guids)

        return recovery_time

    @staticmethod
    def time_stepping_recovery(recovery_results):
        """ Converts results to a time frame. Currently gives results for 16 quarters over 4 year.

        Args:
            recovery_results (pd.DataFrame): Total recovery time of financial delay and other factors from REDi framework.

        Returns:
            pd.DataFrame: Time formatted recovery results.

        """
        time_step = 90 / 7  # a quarter in weeks
        year = 4

        total_time = time_step * np.linspace(0, 4 * year, num=17, endpoint=True)

        # Save guid's for later
        recovery_results_guids = recovery_results['guid']

        # Convert recovery time results to numpy
        samples_n1_n2 = recovery_results.drop(columns=['guid']).to_numpy()

        # Number of guids
        num_buildings = recovery_results.shape[0]
        num_samples = samples_n1_n2.shape[1]
        num_times = len(total_time)

        # Generate a numpy to hold the results as desired
        times_np = np.full((num_buildings, num_times), 1111.0)

        for build in range(0, num_buildings):
            for i in range(len(total_time)):
                fun_state = np.count_nonzero(samples_n1_n2[build, :] < total_time[i]) / num_samples
                times_np[build, i] = np.round(fun_state, 2)

        colnames = [f'quarter_{i}' for i in range(0, num_times)]

        time_stepping_recovery = pd.DataFrame(times_np, columns=colnames)
        time_stepping_recovery.insert(0, 'guid', recovery_results_guids)

        return time_stepping_recovery

    def get_spec(self):
        """Get specifications of the commercial building recovery analysis.

        Returns:
            obj: A JSON object of specifications of the commercial building recovery analysis.

        """
        return {
            'name': 'commercial-building-recovery',
            'description': 'calculate commercial building recovery',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'name of the result',
                    'type': str
                },
                {
                    'id': 'num_samples',
                    'required': True,
                    'description': 'Number of sample scenarios',
                    'type': int
                },
                {
                    'id': 'repair_key',
                    'required': False,
                    'description': 'Repair key to use in mapping dataset',
                    'type': str
                },
                {
                    'id': 'seed',
                    'required': False,
                    'description': 'Initial seed for the probabilistic model',
                    'type': int
                }
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5', 'ergo:buildingInventoryVer6',
                             'ergo:buildingInventoryVer7']
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                },
                {
                    'id': 'sample_damage_states',
                    'required': True,
                    'description': 'Sample damage states',
                    'type': ['incore:sampleDamageState']
                },
                {
                    'id': 'mcs_failure',
                    'required': True,
                    'description': 'mcs_failure',
                    'type': ['incore:failureProbability']
                },
                {
                    'id': 'delay_factors',
                    'required': True,
                    'description': 'Delay impeding factors such as post-disaster inspection, insurance claim, '
                                   'and government permit based on building\'s damage state. Provided by REDi framework',
                    'type': ['incore:buildingRecoveryFactors']
                },
                {
                    'id': 'building_dmg',
                    'required': True,
                    'description': 'damage result that has damage intervals',
                    'type': ['ergo:buildingDamageVer6']
                }
            ],
            'output_datasets': [
                {
                    'id': 'total_delay',
                    'description': 'CSV file of commercial building delay time',
                    'type': 'incore:buildingRecoveryDelay'
                },
                {
                    'id': 'recovery',
                    'description': 'CSV file of commercial building recovery time',
                    'type': 'incore:buildingRecoveryTime'
                },
                {
                    'id': 'time_stepping_recovery',
                    'description': 'CSV file of commercial building recovery percent',
                    'type': 'incore:buildingRecovery'
                }
            ]
        }



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import collections
import concurrent.futures
import pandas as pd

from pyincore import BaseAnalysis, AnalysisUtil, RestorationService


class PipelineRestoration(BaseAnalysis):
    """
    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        self.restorationsvc = RestorationService(incore_client)
        super(PipelineRestoration, self).__init__(incore_client)

    def get_spec(self):
        """Get specifications of the Pipeline Restoration analysis.

        Returns:
            obj: A JSON object of specifications of the pipeline restoration analysis.

        """
        return {
            'name': 'pipeline-restoration',
            'description': 'calculate the restoration times for damaged pipelines',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'name of the result csv dataset',
                    'type': str
                },
                {
                    'id': 'num_cpu',
                    'required': False,
                    'description': 'If using parallel execution, the number of cpus to request',
                    'type': int
                },
                {
                    'id': 'num_available_workers',
                    'required': True,
                    'description': 'Number of available workers to work on the repairs',
                    'type': int
                },
                {
                    'id': 'restoration_key',
                    'required': False,
                    'description': 'restoration key to use in mapping dataset',
                    'type': str
                },
            ],
            'input_datasets': [
                {
                    'id': 'pipeline',
                    'required': True,
                    'description': 'Pipeline Inventory',
                    'type': ['ergo:buriedPipelineTopology', 'ergo:pipeline'],
                },
                {
                    'id': 'pipeline_damage',
                    'required': True,
                    'description': 'pipeline damage results with repairs',
                    'type': ['ergo:pipelineDamageVer2', 'ergo:pipelineDamageVer3']
                },
                {
                    'id': 'dfr3_mapping_set',
                    'required': True,
                    'description': 'DFR3 Mapping Set Object',
                    'type': ['incore:dfr3MappingSet'],
                }
            ],
            'output_datasets': [
                {
                    'id': 'pipeline_restoration',
                    'description': 'CSV file of pipeline restoration times',
                    'type': 'incore:pipelineRestorationVer1'
                }
            ]
        }

    def run(self):
        """Executes pipeline restoration analysis."""

        pipelines_df = self.get_input_dataset("pipeline").get_dataframe_from_shapefile()

        pipeline_dmg = self.get_input_dataset("pipeline_damage").get_csv_reader()
        pipelines_dmg_df = pd.DataFrame(list(pipeline_dmg))

        damage_result = pipelines_dmg_df.merge(pipelines_df, on='guid')
        damage_result = damage_result.to_dict(orient='records')

        user_defined_cpu = 1
        if not self.get_parameter("num_cpu") is None and self.get_parameter(
                "num_cpu") > 0:
            user_defined_cpu = self.get_parameter("num_cpu")

        num_workers = AnalysisUtil.determine_parallelism_locally(self,
                                                                 len(
                                                                     damage_result),
                                                                 user_defined_cpu)

        avg_bulk_input_size = int(len(damage_result) / num_workers)
        inventory_args = []
        count = 0
        inventory_list = damage_result

        while count < len(inventory_list):
            inventory_args.append(inventory_list[count:count + avg_bulk_input_size])
            count += avg_bulk_input_size

        restoration_results = self.pipeline_restoration_concurrent_future(
            self.pipeline_restoration_bulk_input, num_workers,
            inventory_args)
        self.set_result_csv_data("pipeline_restoration",
                                 restoration_results, name=self.get_parameter("result_name"))
        return True

    def pipeline_restoration_concurrent_future(self, function_name,
                                               parallelism, *args):
        """Utilizes concurrent.future module.

        Args:
            function_name (function): The function to be parallelized.
            parallelism (int): Number of workers in parallelization.
            *args: All the arguments in order to pass into parameter function_name.

        Returns:
            list: A list of dictionary with restoration details

        """
        res_output = []
        with concurrent.futures.ProcessPoolExecutor(
                max_workers=parallelism) as executor:
            for res_ret in executor.map(function_name, *args):
                res_output.extend(res_ret)

        return res_output

    def pipeline_restoration_bulk_input(self, damage):
        """Run analysis for pipeline restoration calculation

        Args:
            damage (obj): An output of pipeline damage with repair rate

        Returns:
            restoration_results (list): A list of dictionary restoration times and inventory details

        """
        restoration_results = []
        num_available_workers = self.get_parameter("num_available_workers")

        restoration_key = self.get_parameter("restoration_key")
        if restoration_key is None:
            restoration_key = "Restoration ID Code"

        restoration_sets = self.restorationsvc.match_list_of_dicts(self.get_input_dataset("dfr3_mapping_set"),
                                                                   damage, restoration_key)

        for dmg in damage:
            res = self.restoration_time(dmg, num_available_workers, restoration_sets[dmg['guid']])
            restoration_results.append(res)

        return restoration_results

    @staticmethod
    def restoration_time(dmg, num_available_workers, restoration_set):
        """Calculates restoration time for a single pipeline.

        Args:
            dmg (obj): Pipeline damage analysis output for a single entry.
            num_available_workers (int): Number of available workers working on the repairs.
            restoration_set(obj): Restoration curve(s) to be be used

        Returns:
            dict: A dictionary with id/guid and restoration time, along with some inventory metadata
        """
        res_result = collections.OrderedDict()

        if 'guid' in dmg.keys():
            res_result['guid'] = dmg['guid']
        else:
            res_result['guid'] = 'NA'

        res_result['repair_time'] = restoration_set.calculate_restoration_rates(**{
            "break_rate": float(dmg['breakrate']),
            "leak_rate": float(dmg['leakrate']),
            "pipe_length": dmg['length'],
            "num_workers": num_available_workers})['RT']

        return res_result



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class BuildingUtil:
    """Utility methods for the building damage analysis."""
    DEFAULT_FRAGILITY_KEY = "Non-Retrofit Fragility ID Code"
    DEFAULT_TSUNAMI_HMAX_FRAGILITY_KEY = "Non-Retrofit Inundation Fragility ID Code"
    DEFAULT_TSUNAMI_MMAX_FRAGILITY_KEY = "Non-Retrofit MomentumFlux Fragility ID Code"
    DEFAULT_REPAIR_KEY = "Repair ID Code"
    BLDG_STORIES = "no_stories"
    PROPERTIES = "properties"
    BLDG_PERIOD = "period"
    GROUND_FAILURE_PROB = "groundFailureProb"



# Copyright (c) 2022 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import pandas as pd
from pyincore import BaseAnalysis
from pyincore.utils.dataprocessutil import DataProcessUtil


class CombinedWindWaveSurgeBuildingDamage(BaseAnalysis):
    """ Determines overall building maximum damage state from wind, flood and surge-wave damage
    and uses the maximum damage probabilities from the 3 damages to determine overall damage
    
    Args:
        incore_client (IncoreClient): Service authentication.
    """

    def __init__(self, incore_client):
        super(CombinedWindWaveSurgeBuildingDamage, self).__init__(incore_client)

    def run(self):
        """Executes combined wind, wave, surge building damage analysis."""
        # Read Building wind damage
        wind_damage = self.get_input_dataset("wind_damage").get_dataframe_from_csv()

        # Read Building surge-wave damage
        surge_wave_damage = self.get_input_dataset("surge_wave_damage").get_dataframe_from_csv()

        # Read Building flood damage
        flood_damage = self.get_input_dataset("flood_damage").get_dataframe_from_csv()

        wind_max_damage = DataProcessUtil.get_max_damage_state(wind_damage)
        wind_max_damage.rename(columns={'max_state': 'w_max_ds', 'max_prob': 'w_maxprob'}, inplace=True)

        surge_wave_max_damage = DataProcessUtil.get_max_damage_state(surge_wave_damage)
        surge_wave_max_damage.rename(columns={'max_state': 'sw_max_ds', 'max_prob': 'sw_maxprob'}, inplace=True)

        flood_max_damage = DataProcessUtil.get_max_damage_state(flood_damage)
        flood_max_damage.rename(columns={'max_state': 'f_max_ds', 'max_prob': 'f_maxprob'}, inplace=True)

        combined_output = pd.merge(pd.merge(wind_max_damage, surge_wave_max_damage, on='guid'), flood_max_damage,
                                   on='guid')
        
        # Replace DS strings with integers to find maximum damage state
        replace_vals_int = {'DS_0': 0, 'DS_1': 1, 'DS_2': 2, 'DS_3': 3}
        combined_output = combined_output.apply(lambda x: x.replace(replace_vals_int, regex=True))

        # Find maximum among the max_ds columns
        max_damage_states = ['w_max_ds', 'sw_max_ds', 'f_max_ds']
        max_val = combined_output[max_damage_states].max(axis=1)

        # Add maximum of the max damage states
        combined_output['max_state'] = max_val

        # Replace integers with DS strings
        old_ds_vals = [0, 1, 2, 3]
        new_ds_vals = ['DS_0', 'DS_1', 'DS_2', 'DS_3']

        # Put DS strings back in the final output before storing
        combined_output['w_max_ds'] = combined_output['w_max_ds'].replace(old_ds_vals, new_ds_vals)
        combined_output['sw_max_ds'] = combined_output['sw_max_ds'].replace(old_ds_vals, new_ds_vals)
        combined_output['f_max_ds'] = combined_output['f_max_ds'].replace(old_ds_vals, new_ds_vals)
        combined_output['max_state'] = combined_output['max_state'].replace(old_ds_vals, new_ds_vals)

        # Find combined damage
        combined_bldg_dmg = self.get_combined_damage(wind_damage, surge_wave_damage, flood_damage)

        # Create the result containing the 3 combined damages into a single damage
        self.set_result_csv_data("ds_result", combined_bldg_dmg, self.get_parameter("result_name") + "_combined_dmg",
                                 "dataframe")

        # Create the result dataset
        self.set_result_csv_data("result", combined_output, self.get_parameter("result_name") + "_max_state", "dataframe")

        return True
    
    def get_combined_damage(self, wind_dmg: pd.DataFrame, sw_dmg: pd.DataFrame, flood_dmg: pd.DataFrame):
        """Calculates overall building damage
                Determines the overall building damage probabilities from the 3 hazards by taking the maximum.

                Args:
                    wind_dmg (pd.DataFrame): Table of wind damage for the building inventory
                    sw_dmg (pd.DataFrame): Table of surge-wave damage for the building inventory
                    flood_dmg (pd.DataFrame): Table of flood damage for the building inventory

                Returns:
                    pd.DataFrame: An table of combined damage probabilities for the building inventory

                """
        flood_dmg.rename(columns={'LS_0': 'f_LS_0', 'LS_1': 'f_LS_1', 'LS_2': 'f_LS_2', 'DS_0': 'f_DS_0', 
                                  'DS_1': 'f_DS_1', 'DS_2': 'f_DS_2', 'DS_3': 'f_DS_3', 'haz_expose': 'f_haz_expose'}, 
                         inplace=True)

        sw_dmg.rename(columns={'LS_0': 'sw_LS_0', 'LS_1': 'sw_LS_1', 'LS_2': 'sw_LS_2', 'DS_0': 'sw_DS_0',
                               'DS_1': 'sw_DS_1', 'DS_2': 'sw_DS_2', 'DS_3': 'sw_DS_3', 'haz_expose': 
                                   'sw_haz_expose'}, inplace=True)

        wind_dmg.rename(columns={'LS_0': 'w_LS_0', 'LS_1': 'w_LS_1', 'LS_2': 'w_LS_2', 'DS_0': 'w_DS_0',
                                 'DS_1': 'w_DS_1', 'DS_2': 'w_DS_2', 'DS_3': 'w_DS_3', 'haz_expose': 'w_haz_expose'},
                        inplace=True)

        combined_df = pd.merge(pd.merge(wind_dmg, sw_dmg, on='guid'), flood_dmg, on='guid')

        def find_match(row, col_name):
            max_finder = {
                row['f_DS_3']: 'f_',
                row['w_DS_3']: 'w_',
                row['sw_DS_3']: 'sw_'
            }

            return row[max_finder[max(max_finder.keys())] + col_name]

        combined_df['LS_0'] = combined_df.apply(lambda x: find_match(x, col_name="LS_0"), axis=1)
        combined_df['LS_1'] = combined_df.apply(lambda x: find_match(x, col_name="LS_1"), axis=1)
        combined_df['LS_2'] = combined_df.apply(lambda x: find_match(x, col_name="LS_2"), axis=1)
        combined_df['DS_0'] = combined_df.apply(lambda x: find_match(x, col_name="DS_0"), axis=1)
        combined_df['DS_1'] = combined_df.apply(lambda x: find_match(x, col_name="DS_1"), axis=1)
        combined_df['DS_2'] = combined_df.apply(lambda x: find_match(x, col_name="DS_2"), axis=1)
        combined_df['DS_3'] = combined_df.apply(lambda x: find_match(x, col_name="DS_3"), axis=1)
        combined_df['haz_expose'] = combined_df.apply(lambda x: find_match(x, col_name="haz_expose"), axis=1)

        # Remove extra columns that are no longer needed
        combined_df.drop(['w_LS_0', 'w_LS_1', 'w_LS_2', 'sw_LS_0', 'sw_LS_1', 'sw_LS_2', 'f_LS_0', 'f_LS_1',
                          'f_LS_2', 'w_DS_0', 'w_DS_1', 'w_DS_2', 'w_DS_3', 'sw_DS_0', 'sw_DS_1', 'sw_DS_2', 'sw_DS_3',
                          'f_DS_0', 'f_DS_1', 'f_DS_2', 'f_DS_3', 'w_haz_expose', 'sw_haz_expose', 'f_haz_expose'],
                         axis=1, inplace=True)

        return combined_df

    def get_spec(self):
        """Get specifications of the combined wind, wave, and surge building damage analysis.

        Returns:
            obj: A JSON object of specifications of the combined wind, wave, and surge building damage analysis.

        """
        return {
            'name': 'combined-wind-wave-surge-building-damage',
            'description': 'Combined wind wave and surge building damage analysis',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'result dataset name',
                    'type': str
                },
            ],
            'input_datasets': [
                {
                    'id': 'wind_damage',
                    'required': True,
                    'description': 'Wind damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer6']
                },
                {
                    'id': 'surge_wave_damage',
                    'required': True,
                    'description': 'Surge-wave damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer6']
                },
                {
                    'id': 'flood_damage',
                    'required': True,
                    'description': 'Flood damage result that has damage intervals in it',
                    'type': ['ergo:buildingDamageVer6']
                },

            ],
            'output_datasets': [
                {
                    'id': 'ds_result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of damage states for building structural damage',
                    'type': 'ergo:buildingDamageVer6'
                },
                {
                    'id': 'result',
                    'parent_type': 'buildings',
                    'description': 'CSV file of building maximum damage state',
                    'type': 'incore:maxDamageState'
                }
            ]
        }



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import numpy as np
import pandas as pd

from pyincore import BaseAnalysis
from pyincore.analyses.joplinempiricalrestoration.joplinempirrestor_util import JoplinEmpirRestorUtil


class JoplinEmpiricalRestoration(BaseAnalysis):
    """ Joplin Empirical Restoration Model generates a random realization for the restoration time of
    a building damaged in a tornado event to be restored to a certain functionality level. Functionality
    levels in this model are defined according to Koliou and van de Lindt (2020) and range from Functionality
    Level 4 (FL4, the lowest functionality) to Functionality Level 0 (FL0, full functionality).

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(JoplinEmpiricalRestoration, self).__init__(incore_client)

    def run(self):
        """ Executes Joplin empirical restoration model analysis.

        Returns:
            bool: True if successful, False otherwise.

        """
        # Get seed
        seed_i = self.get_parameter("seed")
        target_fl = self.get_parameter("target_functionality_level")

        result_name = self.get_parameter("result_name")

        # Building dataset
        building_set = self.get_input_dataset("buildings").get_dataframe_from_shapefile()
        # Building damage dataset
        building_dmg = self.get_input_dataset("building_dmg").get_dataframe_from_csv(low_memory=False)
        # Building functionality target level
        building_target_fl = self.get_input_dataset("building_functionality_level")
        if building_target_fl is not None:
            building_target_fl = building_target_fl.get_dataframe_from_csv(low_memory=False)
        else:
            building_target_fl = None

        # merge and filter out archetypes > 5
        building_dmg_all = pd.merge(building_dmg, building_set, how="left", on="guid", copy=True, validate="1:1")
        building_dmg_5 = building_dmg_all[["guid", "archetype", "LS_0", "LS_1", "LS_2", "haz_expose"]].copy()
        building_func_5 = building_dmg_5[building_dmg_all["archetype"] <= 5]

        building_func = building_func_5[["guid", "LS_0", "LS_1", "LS_2", "haz_expose"]].copy()
        building_func["targetFL"] = target_fl

        if building_target_fl is not None:
            building_func = pd.merge(building_func, building_target_fl,
                                     how="left", on="guid", copy=True, validate="1:1")
            # Replace NaN value from targetFL_y with targetFL_x value
            building_func["targetFL"] = building_func["targetFL_y"].fillna(building_func["targetFL_x"])
            # Drop merged columns
            building_func = building_func.drop(["targetFL_x", "targetFL_y"], axis=1)
            building_func = building_func.astype({"targetFL": "int64"})

        initial_func_level, restoration_days = self.get_restoration_days(seed_i, building_func)
        building_func["initialFL"] = initial_func_level
        building_func["restorDays"] = restoration_days

        building_func_fin = building_func[["guid", "initialFL", "targetFL", "restorDays"]]
        csv_source = "dataframe"
        self.set_result_csv_data("result", building_func_fin, result_name, csv_source)

        return True

    def get_restoration_days(self, seed_i, building_func):
        """ Calculates restoration days.

        Args:
            seed_i (int): Seed for random number generator to ensure replication if run as part
                of a stochastic analysis, for example in connection with housing unit allocation analysis.
            building_func (pd.DataFrame): Building damage dataset with guid, limit states, hazard exposure
                and a target level column.

        Returns:
            np.array: Initial functionality level based on damage state
            np.array: Building restoration days.

        """
        fl_coef = JoplinEmpirRestorUtil.FL_COEF

        hazard_value = building_func[["haz_expose"]].to_numpy() != "no"
        hazard_value = hazard_value.flatten()

        bdnp = building_func[["LS_0", "LS_1", "LS_2"]].to_numpy()

        # generate a random number between 0 and 1 and see where in boundaries it locates and use it to assign FL,
        # for each building
        rnd_num = np.random.uniform(0, 1, (len(building_func.index, )))
        bdnp_init = np.zeros(len(building_func.index, )).astype(int)  # first, set all to 0
        bdnp_init = np.where(rnd_num < bdnp[:, 0], 1, bdnp_init)  # if rnd_num < LS_0 set to 1
        bdnp_init = np.where(rnd_num < bdnp[:, 1], 2, bdnp_init)  # if rnd_num < LS_0 set to 2
        bdnp_init = np.where(rnd_num < bdnp[:, 2], 3, bdnp_init)  # if rnd_num < LS_0 set to 3

        bdnp_target = building_func["targetFL"].to_numpy()

        means = fl_coef[bdnp_init, bdnp_target, 0]
        sigmas = fl_coef[bdnp_init, bdnp_target, 1]

        np.random.seed(seed_i)
        rest_days = np.random.lognormal(means, sigmas)

        # only when exposed to hazard, otherwise no damage and restoration = 0
        restoration_days = np.where(hazard_value, rest_days, 0).astype(int)
        # F1-F4 notation
        bdnp_init = bdnp_init + 1
        # F0 where not exposed to hazard
        initial_func_level = np.where(hazard_value, bdnp_init, 0).astype(int)

        # Output of the model is restoration_days
        return initial_func_level, restoration_days

    def get_spec(self):
        """Get specifications of the Joplin empirical restoration analysis.

        Returns:
            obj: A JSON object of specifications of the Joplin empirical restoration analysis.

        """
        return {
            "name": "joplin-empirical-restoration",
            "description": "Values (in days) for the predicted restoration time of the building.",
            "input_parameters": [
                {
                    "id": "result_name",
                    "required": True,
                    "description": "result dataset name",
                    "type": str
                },
                {
                    "id": "target_functionality_level",
                    "required": False,
                    "description": "Target functionality level for all infrastructure",
                    "type": int
                },
                {
                    "id": "seed",
                    "required": False,
                    "description": "Initial seed for the tornado hazard value",
                    "type": int
                }
            ],
            "input_datasets": [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer4', 'ergo:buildingInventoryVer5',
                             'ergo:buildingInventoryVer6', 'ergo:buildingInventoryVer7'],
                },
                {
                    "id": "building_dmg",
                    "required": True,
                    "description": "Building damage results CSV file",
                    "type": ["ergo:buildingDamageVer4",
                             "ergo:buildingDamageVer5",
                             "ergo:buildingDamageVer6",
                             "ergo:buildingInventory",
                             "ergo:nsBuildingInventoryDamage",
                             "ergo:nsBuildingInventoryDamageVer2"]
                },
                {
                    "id": "building_functionality_level",
                    "required": False,
                    "description": "Functionality level per building. The target level defaults "
                                   "to target_functionality_level parameter if building not in the dataset",
                    "type": ["incore:buildingFuncTargetVer1"]
                }
            ],
            "output_datasets": [
                {
                    "id": "result",
                    "parent_type": "buildings",
                    "description": "A dataset containing results (format: CSV) with values (in days) for the predicted "
                                   "restoration time of the building.",
                    "type": "incore:restorationTime"
                }
            ]
        }



# Copyright (c) 2023 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


class InfrastructureSpace(object):
    """
    This class models a geographical space.

    Attributes
    ----------
    id : int
        The id of the space
    cost : float
        The cost of preparing the space for a repair action
    """

    def __init__(self, id, cost):
        self.id = id
        self.cost = cost



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
from pyincore import HazardService
from pyincore.models.hazard.hazard import Hazard
from pyincore.models.hazard.hazarddataset import EarthquakeDataset


class Earthquake(Hazard):

    def __init__(self, metadata):
        super().__init__(metadata)
        self.hazardDatasets = []
        if "hazardDatasets" in metadata:
            for hazardDataset in metadata["hazardDatasets"]:
                self.hazardDatasets.append(EarthquakeDataset(hazardDataset))
        self.hazard_type = "earthquake"

    @classmethod
    def from_hazard_service(cls, id: str, hazard_service: HazardService):
        """Get Hazard from hazard service, get metadata as well.

        Args:
            id (str): ID of the Hazard.
            hazard_service (obj): Hazard service.

        Returns:
            obj: Hazard from Data service.

        """
        metadata = hazard_service.get_earthquake_hazard_metadata(id)
        instance = cls(metadata)
        return instance

    def read_hazard_values(self, payload: list, hazard_service=None, **kwargs):
        """ Retrieve bulk earthquake hazard values either from the Hazard service or read it from local Dataset

        Args:
            payload (list):
            hazard_service (obj): Hazard service.
            kwargs (dict): Keyword arguments.
        Returns:
            obj: Hazard values.

        """
        if self.id and self.id != "" and hazard_service is not None:
            return hazard_service.post_earthquake_hazard_values(self.id, payload, **kwargs)
        else:
            return self.read_local_raster_hazard_values(payload)



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/


import json
import warnings

from pyincore.models.units import Units
from pyincore.dataset import Dataset

warnings.filterwarnings("ignore", "", UserWarning)


class Hazard:
    """Hazard.

    Args:
        metadata (dict): Hazard metadata.

    """

    def __init__(self, metadata):

        self.id = metadata["id"] if "id" in metadata else ""
        self.name = metadata['name'] if "name" in metadata else ""
        self.description = metadata['description'] if "description" in metadata else ""
        self.date = metadata['date'] if "date" in metadata else ""
        self.creator = metadata["creator"] if "creator" in metadata else ""
        self.spaces = metadata["spaces"] if "spaces" in metadata else []
        self.hazard_type = metadata["hazard_type"] if "hazard_type" in metadata else ""
        self.hazardDatasets = []

    @classmethod
    def from_json_str(cls, json_str):
        """Create hazard object from json string.

        Args:
            json_str (str): JSON of the Dataset.

        Returns:
            obj: Hazard

        """
        return cls(json.loads(json_str))

    @classmethod
    def from_json_file(cls, file_path):
        """Get hazard from the file.

        Args:
            file_path (str): json file path that holds the definition of a hazard.

        Returns:
            obj: Hazard

        """
        with open(file_path, "r") as f:
            instance = cls(json.load(f))

        return instance

    def read_local_raster_hazard_values(self, payload: list):
        """ Read local hazard values from raster dataset

        Args:
            payload (list):
        Returns:
            obj: Hazard values.

        """

        response = []

        # match demand types with raster file
        for req in payload:
            hazard_values = []
            for index, req_demand_type in enumerate(req["demands"]):
                match = False
                for hazard_dataset in self.hazardDatasets:
                    if hazard_dataset.dataset is None or not isinstance(hazard_dataset.dataset, Dataset):
                        raise Exception("Hazard dataset is not properly attached to the hazard object.")

                    # TODO Consider how to get the closest period
                    # TODO consider pga, pgv, sd conversions
                    if "sa" in req_demand_type.lower():
                        period = req_demand_type.split(" ")[0]
                    else:
                        period = None

                    # find matching raster file (Dataset) to read value from
                    if req_demand_type.lower() == hazard_dataset.demand_type.lower() or \
                            (hasattr(hazard_dataset, 'period') and period == hazard_dataset.period):
                        raw_raster_value = hazard_dataset.dataset.get_raster_value(
                            x=float(req["loc"].split(",")[1]),
                            y=float(req["loc"].split(",")[0]))

                        if raw_raster_value is None:
                            converted_raster_value = raw_raster_value
                        else:
                            # some basic unit conversion
                            converted_raster_value = Units.convert_hazard(raw_raster_value,
                                                                          original_demand_units=hazard_dataset.demand_units,
                                                                          requested_demand_units=req["units"][index])

                            # compare with threshold (optional)
                            threshold_value = hazard_dataset.threshold_value
                            threshold_unit = hazard_dataset.threshold_unit
                            if threshold_value is not None:
                                converted_threshold_value = Units.convert_hazard(threshold_value,
                                                                                 original_demand_units=threshold_unit,
                                                                                 requested_demand_units=req["units"][index])
                                if converted_raster_value < converted_threshold_value:
                                    converted_raster_value = None

                        hazard_values.append(converted_raster_value)
                        match = True
                        break

                if not match:
                    hazard_values.append(-9999.2)  # invalid demand type

            req.update({"hazardValues": hazard_values})
            response.append(req)

        return response



from pyincore import BaseAnalysis
import pandas as pd
import numpy as np


class CapitalShocks(BaseAnalysis):
    """Capital stock shocks for an individual building is equal to the functionality probability multiplied by value
     of the building. This gives us the capital stock loss in the immediate aftermath of a natural disaster
     for a single building. We aggregate each of these individual losses to their associated economic sector
     to calculate the total capital stock lost for that sector. However, the capital stock shocks that are
     used as inputs into the CGE model are scalars for embodying the percent of capital stock remaining.
     We get this by dividing the total capital stock remaining by the total capital stock
     before the natural disaster."

    Args:
        incore_client (IncoreClient): Service authentication.

    """

    def __init__(self, incore_client):
        super(CapitalShocks, self).__init__(incore_client)

    def get_spec(self):
        return {
            'name': 'Capital Shocks',
            'description': 'Capital Shocks generation for cge models.',
            'input_parameters': [
                {
                    'id': 'result_name',
                    'required': True,
                    'description': 'Result dataset name.',
                    'type': str
                },
            ],
            'input_datasets': [
                {
                    'id': 'buildings',
                    'required': True,
                    'description': 'Building Inventory',
                    'type': ['ergo:buildingInventoryVer5', 'ergo:buildingInventoryVer6',
                             'ergo:buildingInventoryVer7']
                },
                {
                    'id': 'buildings_to_sectors',
                    'required': True,
                    'description': 'Mapping of buildings to economic sectors.',
                    'type': ['incore:buildingsToSectors']
                },
                {
                    'id': 'failure_probability',
                    'required': True,
                    'description': 'Failure probability of buildings.',
                    'type': ['incore:failureProbability']
                }
            ],
            'output_datasets': [
                {
                    'id': 'sector_shocks',
                    'required': True,
                    'description': 'Aggregation of building functionality states to capital shocks per sector',
                    'type': 'incore:capitalShocks'
                }
            ]
        }

    def run(self):
        buildings = self.get_input_dataset("buildings").get_inventory_reader()
        buildings_df = pd.DataFrame(list(buildings))
        failure_probability = self.get_input_dataset("failure_probability").get_dataframe_from_csv()
        buildings_to_sectors = self.get_input_dataset("buildings_to_sectors").get_dataframe_from_csv()

        # drop buildings with no sector
        buildings_to_sectors = buildings_to_sectors[pd.notnull(buildings_to_sectors['sector'])]
        building_inventory = pd.DataFrame.from_records(buildings_df["properties"])
        # drop buildings with no appraisal value
        building_inventory = building_inventory[pd.notnull(building_inventory['appr_bldg'])]
        building_inventory['appr_bldg'] = building_inventory['appr_bldg'].astype(float)
        # drop buildings with no failure probability
        failure_probability = failure_probability[pd.notnull(failure_probability['failure_probability'])]

        inventory_failure = pd.merge(building_inventory, failure_probability, on='guid')
        inventory_failure = pd.merge(inventory_failure, buildings_to_sectors, on='guid')
        inventory_failure['cap_rem'] = inventory_failure.appr_bldg * (
                1 - inventory_failure.failure_probability)

        sectors = buildings_to_sectors.sector.unique()
        sector_shocks = {}
        for sector in sectors:
            sector_values = inventory_failure.loc[(inventory_failure['sector'] == sector)]
            sector_cap = sector_values['cap_rem'].sum()
            sector_total = sector_values['appr_bldg'].sum()
            if sector_total == 0:
                continue
            sector_shock = np.divide(sector_cap, sector_total)
            sector_shocks[sector] = sector_shock

        sector_shocks = pd.DataFrame(sector_shocks.items(), columns=['sector', 'shock'])
        result_name = self.get_parameter("result_name")
        self.set_result_csv_data("sector_shocks", sector_shocks, name=result_name, source="dataframe")

        return True



# Copyright (c) 2021 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import geopandas as gpd

from pyincore import Dataset, DataService


class DatasetUtil:
    @staticmethod
    def join_datasets(geodataset, tabledataset):
        """Join Geopands geodataframe and non-geospatial Dataset using GUID field

        Args:
            geodataset (gpd.Dataset):  pyincore Dataset object with geospatial data
            tabledataset (gpd.Dataset): pyincore Dataset object without geospatial data

        Returns:
            gpd.GeoDataFrame: Geopandas DataFrame object

        """
        gdf = gpd.read_file(geodataset.local_file_path)
        df = tabledataset.get_dataframe_from_csv()
        join_gdf = gdf.set_index("guid").join(df.set_index("guid"))

        return join_gdf

    @staticmethod
    def join_table_dataset_with_source_dataset(dataset, client):
        """Creates geopandas geodataframe by joining table dataset and its source dataset

            Args:
                dataset (Dataset): pyincore dataset object
                client (Client): pyincore service client object

            Returns:
                gpd.Dataset: Geopandas geodataframe object.

        """
        is_source_dataset = False
        source_dataset = None

        # check if the given dataset is table dataset
        if dataset.metadata['format'] != 'table' and dataset.metadata['format'] != 'csv':
            print("The given dataset is not a table dataset")
            return None

        # check if source dataset exists
        try:
            source_dataset = dataset.metadata['sourceDataset']
            is_source_dataset = True
        except Exception:
            print("There is no source dataset for the give table dataset")

        if is_source_dataset:
            # merge dataset and source dataset
            geodataset = Dataset.from_data_service(source_dataset, DataService(client))
            joined_gdf = DatasetUtil.join_datasets(geodataset, dataset)
        else:
            return None

        return joined_gdf



# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/

import pandas as pd
import numpy as np


class PopulationDislocationUtil:

    @staticmethod
    def merge_damage_housing_block(building_dmg: pd.DataFrame, hua_inventory: pd.DataFrame,
                                   block_data: pd.DataFrame):
        """Load CSV files to pandas Dataframes, merge them and drop unused columns.

        Args:
            building_dmg (pd.DataFrame): A building damage file in csv format.
            hua_inventory (pd.DataFrame): A housing unit allocation inventory file in csv format.
            block_data (pd.DataFrame): A block data file in csv format.

        Returns:
            pd.DataFrame: A merged table of all three inputs.

        """
        # sometimes HUA inventory has (empty) damage state columns, drop them
        damage_states = ["DS_0", "DS_1", "DS_2", "DS_3"]
        if set(damage_states).issubset(hua_inventory.columns):
            hua_inventory = hua_inventory.drop(columns=damage_states)

        # first merge hazard with house unit allocation inventory on "guid"
        # note guid can be duplicated in housing unit allocation inventory
        df = pd.merge(building_dmg, hua_inventory, how="right", on="guid", validate="1:m")

        # drop columns in building damage that are not used
        col_drops = ["LS_0", "LS_1", "LS_2", "hazardtype", "meandamage", "mdamagedev"]
        for col_drop in col_drops:
            if col_drop in df.columns:
                df = df.drop(columns=[col_drop])

        # further add block data information to the dataframe. BTW astype("Int64") converts floats to int
        # and keeps NaN. We need the conversions for a merge.
        if "bgid" in df.columns:
            df["bgid"] = df["bgid"].astype("Int64")
        elif "blockid" in df.columns:
            df["bgid"] = df["blockid"].astype("Int64")
        elif "blockid_x" in df.columns:
            df = PopulationDislocationUtil.compare_columns(df, "blockid_x", "blockid_y", True)
            if "blockid_x-blockid_y" in df.columns:
                exit("Column bgid is ambiguous, check the input datasets!")
            else:
                df["bgid"] = df["blockid"].astype("Int64")
        block_data["bgid"] = block_data["bgid"].astype("Int64")

        # outer merge on bgid
        final_df = pd.merge(df, block_data, how="left", on="bgid", validate="m:1")

        return final_df

    @staticmethod
    def get_disl_probability(value_loss: np.array, d_sf: np.array,
                             percent_black_bg: np.array, percent_hisp_bg: np.array):
        """
        Calculate dislocation, the probability of dislocation for the household and population.
        Probability of dislocation Damage factor,
        based on current IN-COREv1 algorithm and Bai et al. 2009 damage factors.

        The following variables are need to predict dislocation using logistic model
        see detailed explanation https://opensource.ncsa.illinois.edu/confluence/
        display/INCORE1/Household+and+Population+Dislocation?
        preview=%2F66224473%2F68289561%2FAlgorithm+3+Logistic.pdf

        Args:
            value_loss (np.array): Value loss.
            d_sf (np.array): 'Dummy' parameter.
            percent_black_bg (np.array): Block group data, percentage of black minority.
            percent_hisp_bg (np.array): Block group data, percentage of hispanic minority.

        Returns:
            np.array: Dislocation probability for the household and population.

        """
        # coefficients for the Logistic regression model
        coefficient = {"beta0": -0.42523,
                       "beta1": 0.02480,
                       "beta2": -0.50166,  # single family coefficient
                       "beta3": -0.01826,  # black block group coefficient
                       "beta4": -0.01198}  # hispanic block group coefficient

        disl_prob = np.zeros_like(d_sf)
        try:
            disl_prob = 1.0 / (1 + np.exp(-1.0 * (coefficient["beta0"] * 1 +
                                                  coefficient["beta1"] * (value_loss * 100) +
                                                  coefficient["beta2"] * d_sf +
                                                  coefficient["beta3"] * percent_black_bg +
                                                  coefficient["beta4"] * percent_hisp_bg)))
        except Exception as e:
            print()
            # raise e

        return disl_prob

    @staticmethod
    def get_random_loss(seed_i: int, df: pd.DataFrame, damage_state: str, size: int):
        """
        Calculates value loss for each structure based on random beta distribution
        Value loss based on damage state is an input to the population dislocation model.

        Args:
            seed_i (int): Seed for random normal to ensure replication if run as part of a stochastic analysis,
                for example in connection with housing unit allocation analysis.
            df (pd.DataFrame): Sata frame that includes the alpha, beta, lower bound, upper bound
                for each required damage state
            damage_state (str): Damage state to calculate value loss for.
            size (int): Size of array to be generated.

        Returns:
            np.array: random distribution of value loss for each structure

        """
        # select upper bound and lower bound from input table
        alpha = df.loc[damage_state, 'alpha']
        beta = df.loc[damage_state, 'beta']
        ub = df.loc[damage_state, 'ub']
        lb = df.loc[damage_state, 'lb']

        # Generate array of random values that follow beta distribution for damage state
        random_generator = np.random.RandomState(seed_i)
        rploss = random_generator.beta(alpha, beta, size) * (ub - lb) + lb

        return rploss

    @staticmethod
    def compare_merges(table1_cols, table2_cols, table_merged):
        """ Compare two lists of columns and run compare columns on columns in both lists.
        It assumes that suffixes are _x and _y

        Args:
            table1_cols (list): columns in table 1
            table2_cols (list): columns in table 2
            table_merged (pd.DataFrame): merged table

            Returns:
                pd.DataFrame: Merged table

        """
        match_column = set(table1_cols).intersection(table2_cols)
        for col in match_column:
            # Compare two columns and marked similarity or rename and drop
            if col + "_x" in table_merged.columns and col + "_y" in table_merged.columns:
                table_merged = PopulationDislocationUtil.compare_columns(table_merged,
                                                                         col + "_x",
                                                                         col + "_y", True)
        return table_merged

    @staticmethod
    def compare_columns(table, col1, col2, drop):
        """Compare two columns. If not equal create Tru/False column, if equal rename one of them
        with the base name and drop the other.

        Args:
            table (pd.DataFrame): Data Frame table
            col1 (str): name of column 1
            col2 (str): name of column 2
            drop (bool): rename and drop column

        Returns:
            pd.DataFrame: Table with True/False column

        """
        # Values in columns match or not, add True/False column
        table.loc[table[col1] == table[col2], col1 + "-" + col2] = True
        table.loc[table[col1] != table[col2], col1 + "-" + col2] = False

        if table[col1].equals(table[col2]):
            col1_base = col1.rsplit("_", 1)[0]
            col2_base = col1.rsplit("_", 1)[0]
            if col1_base == col2_base and drop:
                table[col1_base] = table[col1]
                table = table.drop(columns=[col1, col2, col1 + "-" + col2])

        return table



# Copyright (c) 2019 University of Illinois and others. All rights reserved.
#
# This program and the accompanying materials are made available under the
# terms of the Mozilla Public License v2.0 which accompanies this distribution,
# and is available at https://www.mozilla.org/en-US/MPL/2.0/
import os
import pycodestyle

from pyincore.globals import PYINCORE_ROOT_FOLDER

paths = [
    os.path.join(PYINCORE_ROOT_FOLDER, 'pyincore/'),
    os.path.join(PYINCORE_ROOT_FOLDER, 'tests/'),
    os.path.join(PYINCORE_ROOT_FOLDER, 'recipes/'),
    os.path.join(PYINCORE_ROOT_FOLDER, 'scripts/')
]


def test_conformance(paths=paths):
    """Test that pyIncore conforms to PEP-8."""
    style = pycodestyle.StyleGuide(quiet=False, max_line_length=120)
    result = style.check_files(paths)
    assert result.total_errors == 0


test_conformance()
